/*============================================================================
 *  Dfinition de la fonction
 *   de lecture d'un fichier de maillage IggHexa NUMECA
 *============================================================================*/

/*
                      Code_Saturne version 1.3
                      ------------------------

  This file is part of the Code_Saturne Preprocessor, element of the
  Code_Saturne CFD tool.

  Copyright (C) 1999-2007 EDF S.A., France

  contact: saturne-support@edf.fr

  The Code_Saturne Preprocessor is free software; you can redistribute it
  and/or modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2 of
  the License, or (at your option) any later version.

  The Code_Saturne Preprocessor is distributed in the hope that it will be
  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with the Code_Saturne Preprocessor; if not, write to the
  Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor,
  Boston, MA  02110-1301  USA
*/


/*============================================================================
 *                                 Visibilit
 *============================================================================*/

/*----------------------------------------------------------------------------*
 *  Fichiers `include' librairie standard C ou BFT
 *----------------------------------------------------------------------------*/

#include <assert.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <bft_error.h>
#include <bft_file.h>
#include <bft_mem.h>
#include <bft_printf.h>


/*----------------------------------------------------------------------------*
 *  Fichiers `include' visibles du  paquetage global "Utilitaire"
 *----------------------------------------------------------------------------*/

#include "ecs_chaine_glob.h"
#include "ecs_def.h"
#include "ecs_elt_typ_liste.h"
#include "ecs_tab.h"


/*----------------------------------------------------------------------------*
 *  Fichiers `include' visibles des paquetages visibles
 *----------------------------------------------------------------------------*/

#include "ecs_descr.h"
#include "ecs_champ.h"
#include "ecs_entmail.h"
#include "ecs_maillage.h"


/*----------------------------------------------------------------------------*
 *  Fichiers `include' visibles du  paquetage courant
 *----------------------------------------------------------------------------*/

#include "ecs_entmail_pre.h"


/*----------------------------------------------------------------------------*
 *  Fichier  `include' du  paquetage courant associ au fichier courant
 *----------------------------------------------------------------------------*/

#include "ecs_pre_hex.h"


/*----------------------------------------------------------------------------*
 *  Fichiers `include' privs   du  paquetage courant
 *----------------------------------------------------------------------------*/


/*============================================================================
 *                  Dfinitions de paramtres-macros
 *============================================================================*/

/* Pour une lecture de 80 caracteres par ligne  */
/* auxquels il faut ajouter le `\n' et le `\0'  */
/* pour l'affectation dans la chaine receptrice */

#define ECS_LOC_LNG_MAX_CHAINE_HEX  82           /* Dimension des chaines */


enum {
  X ,
  Y ,
  Z
} ;


#define ECS_LOC_PRODUIT_VECTORIEL(prod_vect, vect1, vect2) ( \
prod_vect[X] = vect1[Y] * vect2[Z] - vect2[Y] * vect1[Z] ,   \
prod_vect[Y] = vect2[X] * vect1[Z] - vect1[X] * vect2[Z] ,   \
prod_vect[Z] = vect1[X] * vect2[Y] - vect2[X] * vect1[Y]   )


#define ECS_LOC_PRODUIT_SCALAIRE(vect1, vect2)                        ( \
  vect1[X] * vect2[X] + vect1[Y] * vect2[Y] + vect1[Z] * vect2[Z] )


/*============================================================================
 *                       Prototypes de fonctions prives
 *============================================================================*/

/*----------------------------------------------------------------------------*
 *                      Lecture de la table de connectivit
 *----------------------------------------------------------------------------*/

static void ecs_loc_pre_hex__lit_elements
(
 bft_file_t     *const   fic_maillage    , /* --> Descripteur fichier  lire  */
 ecs_int_t      *const   num_ligne       , /* <-> Compteur des lignes lues    */
 ecs_int_t      *const   nbr_som         , /* <-- Nombre de sommets           */
 ecs_int_t      *const   nbr_are         , /* <-- Nombre d'artes lues        */
 ecs_int_t      *const   nbr_fac         , /* <-- Nombre de faces lues        */
 ecs_int_t      *const   nbr_cel         , /* <-- Nombre de cellules lues     */
 ecs_real_t   * *const som_val_coord     , /* <-- Coordonnes des sommets     */
 ecs_int_t    * *const are_val_som       , /* <-- Connectivit sommets/artes */
 ecs_int_t    * *const fac_val_are       , /* <-- Connectivit artes/faces   */
 ecs_int_t    * *const cel_val_fac       , /* <-- Connectivit faces/cellules */
 ecs_int_t    * *const fac_val_hierarch  , /* <-- Hirarchie des faces        */
 ecs_int_t    * *const cel_val_hierarch    /* <-- Hirarchie des cellules     */
) ;


/*----------------------------------------------------------------------------*
 *                      Lecture des liens avec la CAO (rfrences)
 *----------------------------------------------------------------------------*/

static void ecs_loc_pre_hex__lit_references
(
 bft_file_t     *const   fic_maillage    , /* --> Descripteur fichier  lire  */
 ecs_int_t      *const   num_ligne       , /* <-> Compteur des lignes lues    */
 ecs_int_t      *const   nbr_fac_cao     , /* <-- Nbr. faces dans la CAO      */
 ecs_int_t    * *const   nbr_fac_fac_cao , /* <-- Nbr. faces par face CAO     */
 ecs_int_t  * * *const   lis_fac_fac_cao   /* <-- Liste faces par face CAO    */
) ;


/*----------------------------------------------------------------------------*
 *    Affectation des rfrences des faces  la CAO aux faces du maillage
 *----------------------------------------------------------------------------*/

static ecs_int_t * ecs_loc_pre_hex__prep_reference
(
 ecs_int_t       const nbr_fac           , /* --> Nombre de faces             */
 ecs_int_t      *const fac_val_hierarch  , /* --> Hirarchie des faces        */
 ecs_int_t      *const nbr_fac_cao       , /* <-- Nbr. faces dans la CAO      */
 ecs_int_t    * *const nbr_fac_fac_cao   , /* <-- Nbr. faces par face CAO     */
 ecs_int_t  * * *const lis_fac_fac_cao     /* <-- Liste faces par face CAO    */
) ;


/*----------------------------------------------------------------------------*
 *    Suppression des cellules hirarchiques et compactage associ
 *----------------------------------------------------------------------------*/

static void ecs_loc_pre_hex__compacte
(
 ecs_int_t      *const nbr_som           , /* --> Nombre de sommets           */
 ecs_int_t      *const nbr_are           , /* --> Nombre d'artes             */
 ecs_int_t      *const nbr_fac           , /* --> Nombre de faces             */
 ecs_int_t      *const nbr_cel           , /* <-> Nombre de cellules          */
 ecs_real_t   * *const som_val_coord     , /* --> Coordonnes des sommets     */
 ecs_int_t    * *const are_val_som       , /* --> Connectivit sommets/artes */
 ecs_int_t    * *const fac_val_are       , /* --> Connectivit artes/faces   */
 ecs_int_t    * *const cel_val_fac       , /* --> Connectivit faces/cellules */
 ecs_int_t    * *const fac_val_hierarch  , /* --> Hirarchie des faces        */
 ecs_int_t    * *const cel_val_hierarch  , /* --> Hirarchie des cellules     */
 ecs_int_t    * *const fac_val_reference   /* --> Rfrences des faces        */
) ;


/*----------------------------------------------------------------------------*
 *    Construction de la connectivit nodale
 *----------------------------------------------------------------------------*/

static void ecs_loc_pre_hex__connect_nod
(
 ecs_int_t      *const nbr_are           , /* <-> Nombre d'artes             */
 ecs_int_t      *const nbr_fac           , /* <-> Nombre de faces             */
 ecs_int_t             nbr_cel           , /* --> Nombre de cellules          */
 ecs_real_t     *const som_val_coord     , /* --> Coordonnes des sommets     */
 ecs_int_t    * *const are_val_som       , /* <-> Connectivit sommets/artes */
 ecs_int_t    * *const fac_val_are       , /* <-> Connectivit artes/faces   */
 ecs_int_t    * *const fac_val_som       , /* <-- Connectivit sommets/faces  */
 ecs_int_t    * *const cel_val_fac       , /* <-> Connectivit faces/cellules */
 ecs_int_t    * *const cel_val_som       , /* <-- Connectivit sommets/cel.   */
 ecs_int_t    * *const fac_val_hierarch  , /* <-> Hirarchie des faces        */
 ecs_int_t    * *const fac_val_reference   /* <-> Rfrences des faces        */
) ;


/*----------------------------------------------------------------------------*
 *    Construction de la structure maillage
 *----------------------------------------------------------------------------*/

static ecs_entmail_t ** ecs_loc_pre_hex__cree_entmail
(
 ecs_int_t             nbr_som           , /* --> Nombre de sommets           */
 ecs_int_t             nbr_fac           , /* --> Nombre de faces             */
 ecs_int_t             nbr_cel           , /*  -> Nombre de cellules          */
 ecs_real_t   * *const som_val_coord     , /* --> Coordonnes des sommets     */
 ecs_int_t    * *const are_val_som       , /* <-> Connectivit sommets/artes */
 ecs_int_t    * *const fac_val_som       , /* <-- Connectivit sommets/faces  */
 ecs_int_t    * *const cel_val_som       , /* <-- Connectivit sommets/cel.   */
 ecs_int_t    * *const fac_val_reference   /* <-> Rfrences des faces        */
) ;


/*----------------------------------------------------------------------------
 *   Transformation d'une hirarchie en filiation et affectation  une entit
 *----------------------------------------------------------------------------*/

static void ecs_loc_pre_hex__cree_filiation
(
 ecs_int_t               nbr_elt           , /* --> Nombre d'lments         */
 ecs_int_t      * *const ele_val_hierarch  , /* <-> Hirarchie des lments   */
 ecs_entmail_t    *const entmail             /* <-> Entit de maillage        */
) ;


/*----------------------------------------------------------------------------
 *   Lecture d'un entier au format " %d" sur fichier ASCII avec des lignes
 *  potentiellement trs longues. On lit alors une partie de la ligne
 *  dans un tampon, dont on fait glisser le contenu aprs extraction de
 *  l'entier.
 *----------------------------------------------------------------------------*/

static int ecs_loc_pre_hex__scan_d_ligne
(
 bft_file_t       *const fic               , /* --> Fichier                   */
 ecs_int_t        *const num_ligne         , /* <-> Numro de ligne           */
 char             *const chaine            , /* <-> Tampon associ            */
 ecs_int_t               len                 /* <-> Longueur de la chane     */
) ;


/*============================================================================
 *                             Fonctions publiques
 *============================================================================*/

/*----------------------------------------------------------------------------*
 *  Lecture d'un fichier Hex NUMECA
 *   et affectation des donnees dans la structure de maillage
 *----------------------------------------------------------------------------*/

ecs_maillage_t * ecs_pre_hex__lit_maillage
(                                       /* <-- Renvoie un pointeur sur        */
                                        /*      une structure de maillage     */
 const char *const nom_fic_maillage     /* --> Nom du fichier a lire          */
)
{
  bft_file_t      *fic_maillage        ; /* Descripteur du fichier            */
  bft_file_type_t  type_fic_maillage   ; /* Type du fichier de maillage       */
  ecs_int_t        retour              ; /* Retour fnctions `scan' pour test  */
  char             chaine[ECS_LOC_LNG_MAX_CHAINE_HEX] ;        /* Ligne lue   */
  ecs_int_t        num_ligne           ; /* Compteur des lignes lues          */
  ecs_maillage_t  *maillage            ; /* Structure de maillage             */

  ecs_int_t        nbr_som ;                   /* Nombre de sommets           */
  ecs_int_t        nbr_are ;                   /* Nombre d'artes             */
  ecs_int_t        nbr_fac ;                   /* Nombre de faces             */
  ecs_int_t        nbr_cel ;                   /* Nombre de cellules          */
  ecs_real_t      *som_val_coord ;             /* Coordonnes des sommets     */
  ecs_int_t       *are_val_som ;               /* Connectivit sommets/artes */
  ecs_int_t       *fac_val_are ;               /* Connectivit artes/faces   */
  ecs_int_t       *fac_val_som ;               /* Connectivit sommets/faces  */
  ecs_int_t       *cel_val_fac ;               /* Connectivit faces/cellules */
  ecs_int_t       *cel_val_som ;               /* Connectivit sommets/cel.   */
  ecs_int_t       *fac_val_hierarch ;          /* Hirarchie des faces        */
  ecs_int_t       *cel_val_hierarch ;          /* Hirarchie des cellules     */
  ecs_int_t       *fac_val_reference ;         /* Rfrences des faces        */

  ecs_int_t        nbr_fac_cao ;
  ecs_int_t       *nbr_fac_fac_cao ;
  ecs_int_t      **lis_fac_fac_cao ;

  ecs_entmail_t  **vect_entmail ;

  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  /* Affichage du titre */
  /*====================*/

  bft_printf(_("\n\n"
               "Reading mesh from file in NUMECA Hex format\n"
               "----------------------\n")) ;

  bft_printf(_("  Mesh file: %s\n\n"),
             nom_fic_maillage) ;


  /* Initialisations */
  /*=================*/

  num_ligne = 1 ;


  /* Ouverture du fichier Hex en lecture */
  /*---------------------------------------*/

  fic_maillage = bft_file_open(nom_fic_maillage,
                               BFT_FILE_MODE_READ,
                               BFT_FILE_TYPE_BINARY) ;

  bft_file_set_big_endian(fic_maillage) ;


  /* Lecture de l'entte */

  {
    char          *header;     /* Entte */
    ecs_int_32_t   headersize; /* Taille de l'entte */

    char * version ;
    char * fileversion ;
    char * tmpdate ;


    bft_file_read(chaine, sizeof(char), 2, fic_maillage) ;

    if (chaine[0] == '0') {

      bft_file_free(fic_maillage) ;

      fic_maillage = bft_file_open(nom_fic_maillage,
                                   BFT_FILE_MODE_READ,
                                   BFT_FILE_TYPE_TEXT) ;
      bft_file_read(chaine, sizeof(char), 2, fic_maillage) ;

    }

    type_fic_maillage = bft_file_get_type(fic_maillage) ;

    if (type_fic_maillage == BFT_FILE_TYPE_BINARY)
      bft_file_read(&headersize, sizeof(ecs_int_32_t), 1, fic_maillage) ;
    else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {
      bft_file_read(chaine, sizeof(char), 4, fic_maillage) ;
      sscanf(chaine, "%4d", &headersize) ;
    }

    BFT_MALLOC(header, headersize + 1, char);
    BFT_MALLOC(version, headersize + 1, char);
    BFT_MALLOC(fileversion, headersize + 1, char);
    BFT_MALLOC(tmpdate, headersize + 1, char);

    if (type_fic_maillage == BFT_FILE_TYPE_BINARY)
      bft_file_read(header, sizeof(char), headersize, fic_maillage) ;
    else if (type_fic_maillage == BFT_FILE_TYPE_TEXT)
      bft_file_gets(header, headersize, fic_maillage, &num_ligne) ;

    retour = sscanf(header, "%*s %*s %*s %*s %s %*s %*s %s %s",
                    version, fileversion, tmpdate) ;
    if (retour != 3)
      bft_error(__FILE__, __LINE__, errno,
                _("Error reading line %d of file \"%s\"."),
                1, nom_fic_maillage);

    bft_printf(_("  IggHexa %s, format version  %s, %s\n\n"),
               version, fileversion, tmpdate);

    BFT_FREE(header);
    BFT_FREE(version);
    BFT_FREE(fileversion);
    BFT_FREE(tmpdate);


    if (type_fic_maillage == BFT_FILE_TYPE_BINARY)
      bft_file_read(chaine, sizeof(char), 1, fic_maillage) ;
    else if (type_fic_maillage == BFT_FILE_TYPE_TEXT)
      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, &num_ligne) ;

    assert (chaine[0] == '0' || chaine[0] == '1') ;

    if (chaine[0] == '0')
      bft_error(__FILE__, __LINE__, 0,
                _("The IggHexa file %s\ncontains no mesh"),
                nom_fic_maillage);

  }


  /* Lecture des noeuds et de la connectivit */
  /*------------------------------------------*/

  ecs_loc_pre_hex__lit_elements(fic_maillage,
                                &num_ligne,
                                &nbr_som,
                                &nbr_are,
                                &nbr_fac,
                                &nbr_cel,
                                &som_val_coord,
                                &are_val_som,
                                &fac_val_are,
                                &cel_val_fac,
                                &fac_val_hierarch,
                                &cel_val_hierarch);


  /*
    Lecture de l'indicateur prcisant si l'on a un lien avec la CAO
    (et donc des rfrences de faces et cellules)
  */

  if (type_fic_maillage == BFT_FILE_TYPE_BINARY)
    bft_file_read(chaine, sizeof(char), 1, fic_maillage) ;
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT)
    bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                  fic_maillage, &num_ligne) ;

  assert (chaine[0] == '0' || chaine[0] == '1') ;

  if (chaine[0] == '0') {
    ecs_warn() ;
    bft_printf(_("The IggHexa file %s\n"
                 "contains no links with CAD.\n"
                 "No face will thus be referenced"),
               nom_fic_maillage);
  }


  /* Lecture des liens avec la CAO */
  /* ----------------------------- */

  ecs_loc_pre_hex__lit_references(fic_maillage,
                                  &num_ligne,
                                  &nbr_fac_cao,
                                  &nbr_fac_fac_cao,
                                  &lis_fac_fac_cao);


  /* Fermeture du fichier de lecture du maillage */
  /*---------------------------------------------*/

  bft_file_free(fic_maillage) ;


  /* Prparation des rfrences et libration des liens avec la CAO */
  /*----------------------------------------------------------------*/

  fac_val_reference
    = ecs_loc_pre_hex__prep_reference(nbr_fac,
                                      fac_val_hierarch,
                                      &nbr_fac_cao,
                                      &nbr_fac_fac_cao,
                                      &lis_fac_fac_cao);


  /* Prparation du maillage */
  /*-------------------------*/

  ecs_loc_pre_hex__compacte(&nbr_som,
                            &nbr_are,
                            &nbr_fac,
                            &nbr_cel,
                            &som_val_coord,
                            &are_val_som,
                            &fac_val_are,
                            &cel_val_fac,
                            &fac_val_hierarch,
                            &cel_val_hierarch,
                            &fac_val_reference);

  ecs_loc_pre_hex__connect_nod(&nbr_are,
                               &nbr_fac,
                               nbr_cel,
                               som_val_coord,
                               &are_val_som,
                               &fac_val_are,
                               &fac_val_som,
                               &cel_val_fac,
                               &cel_val_som,
                               &fac_val_hierarch,
                               &fac_val_reference);

  vect_entmail = ecs_loc_pre_hex__cree_entmail(nbr_som,
                                               nbr_fac,
                                               nbr_cel,
                                               &som_val_coord,
                                               &are_val_som,
                                               &fac_val_som,
                                               &cel_val_som,
                                               &fac_val_reference);

  ecs_loc_pre_hex__cree_filiation(nbr_fac,
                                  &fac_val_hierarch,
                                  vect_entmail[ECS_ENTMAIL_FAC]) ;


  /* Cration de la structure maillage associe */

  maillage = ecs_maillage__cree_nodal(ECS_DIM_3,
                                      vect_entmail) ;

  BFT_FREE(vect_entmail) ;

  return maillage ;


}


/*============================================================================
 *                              Fonctions prives
 *============================================================================*/

/*----------------------------------------------------------------------------*
 *                      Lecture de la table de connectivit
 *----------------------------------------------------------------------------*/

static void ecs_loc_pre_hex__lit_elements
(
 bft_file_t     *const fic_maillage     ,  /* --> Descripteur fichier  lire  */
 ecs_int_t      *const num_ligne        ,  /* <-> Compteur des lignes lues    */
 ecs_int_t      *const nbr_som          ,  /* <-- Nombre de sommets           */
 ecs_int_t      *const nbr_are          ,  /* <-- Nombre d'artes lues        */
 ecs_int_t      *const nbr_fac          ,  /* <-- Nombre de faces lues        */
 ecs_int_t      *const nbr_cel          ,  /* <-- Nombre de cellules lues     */
 ecs_real_t   * *const som_val_coord    ,  /* <-- Coordonnes des sommets     */
 ecs_int_t    * *const are_val_som      ,  /* <-- Connectivit sommets/artes */
 ecs_int_t    * *const fac_val_are      ,  /* <-- Connectivit artes/faces   */
 ecs_int_t    * *const cel_val_fac      ,  /* <-- Connectivit faces/cellules */
 ecs_int_t    * *const fac_val_hierarch ,  /* <-- Hirarchie des faces        */
 ecs_int_t    * *const cel_val_hierarch    /* <-- Hirarchie des cellules     */
)
{

  ecs_int_t      retour ;              /* Retour fonctions `scan' pour test   */
  char           chaine[ECS_LOC_LNG_MAX_CHAINE_HEX] ;          /* Ligne lue   */

  ecs_int_t      cpt_elt = 0 ;         /* Compteur des lments lus           */
  ecs_int_32_t   nbr_elt_lu ;

  bft_file_type_t  type_fic_maillage ; /* Type du fichier de maillage         */

  /* Variables Hex lues */

  ecs_int_32_t  *are_val_som_lu ;
  ecs_int_32_t  *fac_val_are_lu ;
  ecs_int_32_t  *cel_val_fac_lu ;

  ecs_int_32_t  *are_val_hierarch_lu ;
  ecs_int_32_t  *fac_val_hierarch_lu ;
  ecs_int_32_t  *cel_val_hierarch_lu ;

  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  /* Type de maillage */

  type_fic_maillage = bft_file_get_type(fic_maillage) ;


  /* Sommets */
  /*=========*/

  /*
    Pour assurer la portabilit de lecture binaire (double ne change que
    trs rarement avec l'architecture,  la diffrence de int)
  */

  assert (sizeof(ecs_real_t) == sizeof(double)) ;


  /* Nombre de sommets */

  if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {

    bft_file_read(&nbr_elt_lu, sizeof(ecs_int_32_t), 1, fic_maillage) ;
    *nbr_som = (ecs_int_t) nbr_elt_lu ;

  }
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {

    bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                  fic_maillage, num_ligne) ;
    retour = sscanf(chaine, " %d", nbr_som);
    if (retour != 1)
      bft_error(__FILE__, __LINE__, errno,
                _("Error reading line %d of file \"%s\"."),
                *num_ligne, bft_file_get_name(fic_maillage));

  }


  /* Dfinition des sommets */

  BFT_MALLOC((*som_val_coord), *nbr_som * ECS_DIM_3, ecs_real_t);


  if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {

    bft_file_read(*som_val_coord, sizeof(ecs_real_t),
                  *nbr_som * ECS_DIM_3, fic_maillage) ;

  }
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {

    static char format_float[] = "%*d %g %g %g";
    static char format_double[] = "%*d %lg %lg %lg";

    char *format_lec ;

    if (sizeof(ecs_real_t) == sizeof(float))
      format_lec = format_float ;
    else if (sizeof(ecs_real_t) == sizeof(double))
      format_lec = format_double ;
    else
      assert (   sizeof(ecs_real_t) == sizeof(float)
              || sizeof(ecs_real_t) == sizeof(double)) ;


    for (cpt_elt = 0 ;
         cpt_elt < *nbr_som ;
         cpt_elt++) {

      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;

      retour = sscanf(chaine, format_lec,
                      &((*som_val_coord)[cpt_elt * ECS_DIM_3    ]),
                      &((*som_val_coord)[cpt_elt * ECS_DIM_3 + 1]),
                      &((*som_val_coord)[cpt_elt * ECS_DIM_3 + 2]));
      if (retour != 3)
        bft_error(__FILE__, __LINE__, errno,
                  _("Error reading line %d of file \"%s\"."),
                  *num_ligne, bft_file_get_name(fic_maillage));

    }

  }


#if 0 && defined(DEBUG) && !defined(NDEBUG)

  for (cpt_elt = 0 ; cpt_elt < *nbr_som ; cpt_elt++)
    printf("som  %d : %g %g %g\n", cpt_elt,
           (*som_val_coord)[cpt_elt * ECS_DIM_3    ],
           (*som_val_coord)[cpt_elt * ECS_DIM_3 + 1],
           (*som_val_coord)[cpt_elt * ECS_DIM_3 + 2]);

#endif

  /* Artes */
  /*========*/


  /* Nombre d'artes */

  if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {

    bft_file_read(&nbr_elt_lu, sizeof(ecs_int_32_t), 1, fic_maillage) ;
    *nbr_are = (ecs_int_t) nbr_elt_lu ;

  }
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {

    bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                  fic_maillage, num_ligne) ;
    retour = sscanf(chaine, " %d", nbr_are);
    if (retour != 1)
      bft_error(__FILE__, __LINE__, errno,
                _("Error reading line %d of file \"%s\"."),
                *num_ligne, bft_file_get_name(fic_maillage));

  }

  /* Dfinition des artes */

  BFT_MALLOC(are_val_som_lu, *nbr_are * 2, ecs_int_32_t);

  cpt_elt = 0 ;

  if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {

    bft_file_read(are_val_som_lu, sizeof(ecs_int_32_t),
                  *nbr_are * 2, fic_maillage) ;

  }
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {

    for (cpt_elt = 0 ;
         cpt_elt < *nbr_are ;
         cpt_elt++) {

      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;

      retour = sscanf(chaine, "%*d %d %d",
                      &(are_val_som_lu[cpt_elt * 2    ]),
                      &(are_val_som_lu[cpt_elt * 2 + 1]));
      if (retour != 2)
        bft_error(__FILE__, __LINE__, errno,
                  _("Error reading line %d of file \"%s\"."),
                  *num_ligne, bft_file_get_name(fic_maillage));

    }

  }

  if (sizeof(ecs_int_t) == sizeof(ecs_int_32_t)) {
    (*are_val_som) = (ecs_int_t *) are_val_som_lu ;
    are_val_som_lu = NULL ;
  }
  else {
    ecs_int_t ind ;
    BFT_MALLOC((*are_val_som), *nbr_are * 2, ecs_int_t);
    for (ind = 0 ; ind < *nbr_are * 2; ind++)
      (*are_val_som)[ind] = (ecs_int_t) are_val_som_lu[ind] ;
    BFT_FREE(are_val_som_lu) ;
  }

#if 0 && defined(DEBUG) && !defined(NDEBUG)

  for (cpt_elt = 0 ; cpt_elt < *nbr_are ; cpt_elt++)
    printf("are  %d : sommets %d et %d\n", cpt_elt,
           (*are_val_som)[cpt_elt * 2    ],
           (*are_val_som)[cpt_elt * 2 + 1]);

#endif

  /* Hirarchie des artes */

  BFT_MALLOC(are_val_hierarch_lu, *nbr_are * 2, ecs_int_32_t);

  cpt_elt = 0 ;

  if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {

    bft_file_read(are_val_hierarch_lu, sizeof(ecs_int_32_t),
                  *nbr_are * 2, fic_maillage) ;

  }
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {

    for (cpt_elt = 0 ;
         cpt_elt < *nbr_are ;
         cpt_elt++) {

      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;

      retour = sscanf(chaine, "%*d %d %d",
                      &(are_val_hierarch_lu[cpt_elt * 2    ]),
                      &(are_val_hierarch_lu[cpt_elt * 2 + 1]));
      if (retour != 2)
        bft_error(__FILE__, __LINE__, errno,
                  _("Error reading line %d of file \"%s\"."),
                  *num_ligne, bft_file_get_name(fic_maillage));

    }

  }

  BFT_FREE(are_val_hierarch_lu) ;


  /* Faces */
  /*=======*/


  /* Nombre de faces */

  if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {

    bft_file_read(&nbr_elt_lu, sizeof(ecs_int_32_t), 1, fic_maillage) ;
    *nbr_fac = (ecs_int_t) nbr_elt_lu ;

  }
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {

    bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                  fic_maillage, num_ligne) ;
    retour = sscanf(chaine, " %d", nbr_fac);
    if (retour != 1)
      bft_error(__FILE__, __LINE__, errno,
                _("Error reading line %d of file \"%s\"."),
                *num_ligne, bft_file_get_name(fic_maillage));

  }

  /* Dfinition des faces */

  BFT_MALLOC(fac_val_are_lu, *nbr_fac * 4, ecs_int_32_t);

  cpt_elt = 0 ;

  if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {

    bft_file_read(fac_val_are_lu, sizeof(ecs_int_32_t),
                  *nbr_fac * 4, fic_maillage) ;

  }
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {

    for (cpt_elt = 0 ;
         cpt_elt < *nbr_fac ;
         cpt_elt++) {

      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;

      retour = sscanf(chaine, "%*d %d %d %d %d",
                      &(fac_val_are_lu[cpt_elt * 4    ]),
                      &(fac_val_are_lu[cpt_elt * 4 + 1]),
                      &(fac_val_are_lu[cpt_elt * 4 + 2]),
                      &(fac_val_are_lu[cpt_elt * 4 + 3]));
      if (retour != 4)
        bft_error(__FILE__, __LINE__, errno,
                  _("Error reading line %d of file \"%s\"."),
                  *num_ligne, bft_file_get_name(fic_maillage));

    }

  }

  if (sizeof(ecs_int_t) == sizeof(ecs_int_32_t)) {
    (*fac_val_are) = (ecs_int_t *) fac_val_are_lu ;
    fac_val_are_lu = NULL ;
  }
  else {
    ecs_int_t ind ;
    BFT_MALLOC((*fac_val_are), *nbr_fac * 4, ecs_int_t);
    for (ind = 0 ; ind < *nbr_fac * 4 ; ind++)
      (*fac_val_are)[ind] = (ecs_int_t) fac_val_are_lu[ind] ;
    BFT_FREE(fac_val_are_lu) ;
  }

#if 0 && defined(DEBUG) && !defined(NDEBUG)

  for (cpt_elt = 0 ; cpt_elt < *nbr_fac ; cpt_elt++)
    printf("fac  %d : artes %d, %d, %d et %d\n", cpt_elt,
           (*fac_val_are)[cpt_elt*4    ], (*fac_val_are)[cpt_elt*4 + 1],
           (*fac_val_are)[cpt_elt*4 + 2], (*fac_val_are)[cpt_elt*4 + 3]);

#endif

  /* Hirarchie des faces */


  BFT_MALLOC(fac_val_hierarch_lu, *nbr_fac * 2, ecs_int_32_t);

  cpt_elt = 0 ;

  if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {

    bft_file_read(fac_val_hierarch_lu, sizeof(ecs_int_32_t),
                  *nbr_fac * 2, fic_maillage) ;

  }
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {

    for (cpt_elt = 0 ;
         cpt_elt < *nbr_fac ;
         cpt_elt++) {

      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;

      retour = sscanf(chaine, "%*d %d %d",
                      &(fac_val_hierarch_lu[cpt_elt * 2    ]),
                      &(fac_val_hierarch_lu[cpt_elt * 2 + 1]));
      if (retour != 2)
        bft_error(__FILE__, __LINE__, errno,
                  _("Error reading line %d of file \"%s\"."),
                  *num_ligne, bft_file_get_name(fic_maillage));

    }

  }

  if (sizeof(ecs_int_t) == sizeof(ecs_int_32_t)) {
    (*fac_val_hierarch) = (ecs_int_t *) fac_val_hierarch_lu ;
    fac_val_hierarch_lu = NULL ;
  }
  else {
    ecs_int_t ind ;
    BFT_MALLOC((*fac_val_hierarch), *nbr_fac * 2, ecs_int_t);
    for (ind = 0 ; ind < *nbr_fac * 2 ; ind++)
      (*fac_val_hierarch)[ind] = (ecs_int_t) fac_val_hierarch_lu[ind] ;
    BFT_FREE(fac_val_hierarch_lu) ;
  }

#if 0 && defined(DEBUG) && !defined(NDEBUG)

  for (cpt_elt = 0 ; cpt_elt < *nbr_fac ; cpt_elt++)
    printf("fac %d : hierarch %d %d\n", cpt_elt,
           (*fac_val_hierarch)[cpt_elt * 2    ],
           (*fac_val_hierarch)[cpt_elt * 2 + 1]);

#endif


  /* Cellules */
  /*==========*/


  /* Nombre de cellules */

  if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {

    bft_file_read(&nbr_elt_lu, sizeof(ecs_int_32_t), 1, fic_maillage) ;
    *nbr_cel = (ecs_int_t) nbr_elt_lu ;

  }
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {

    bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                  fic_maillage, num_ligne) ;
    retour = sscanf(chaine, " %d", nbr_cel);
    if (retour != 1)
      bft_error(__FILE__, __LINE__, errno,
                _("Error reading line %d of file \"%s\"."),
                *num_ligne, bft_file_get_name(fic_maillage));

  }

  /* Dfinition des cellules */

  BFT_MALLOC(cel_val_fac_lu, *nbr_cel * 6, ecs_int_32_t);

  cpt_elt = 0 ;

  if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {

    bft_file_read(cel_val_fac_lu, sizeof(ecs_int_32_t),
                  *nbr_cel * 6, fic_maillage) ;

  }
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {

    for (cpt_elt = 0 ;
         cpt_elt < *nbr_cel ;
         cpt_elt++) {

      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;

      retour = sscanf(chaine, "%*d %d %d %d %d %d %d",
                      &(cel_val_fac_lu[cpt_elt * 6    ]),
                      &(cel_val_fac_lu[cpt_elt * 6 + 1]),
                      &(cel_val_fac_lu[cpt_elt * 6 + 2]),
                      &(cel_val_fac_lu[cpt_elt * 6 + 3]),
                      &(cel_val_fac_lu[cpt_elt * 6 + 4]),
                      &(cel_val_fac_lu[cpt_elt * 6 + 5]));
      if (retour != 6)
        bft_error(__FILE__, __LINE__, errno,
                  _("Error reading line %d of file \"%s\"."),
                  *num_ligne, bft_file_get_name(fic_maillage));

    }

  }

  if (sizeof(ecs_int_t) == sizeof(ecs_int_32_t)) {
    (*cel_val_fac) = (ecs_int_t *) cel_val_fac_lu ;
    cel_val_fac_lu = NULL ;
  }
  else {
    ecs_int_t ind ;
    BFT_MALLOC((*cel_val_fac), *nbr_cel * 6, ecs_int_t);
    for (ind = 0 ; ind < *nbr_cel * 6 ; ind++)
      (*cel_val_fac)[ind] = (ecs_int_t) cel_val_fac_lu[ind] ;
    BFT_FREE(cel_val_fac_lu) ;
  }

#if 0 && defined(DEBUG) && !defined(NDEBUG)

  for (cpt_elt = 0 ; cpt_elt < *nbr_cel ; cpt_elt++)
    printf("cel  %d : faces %d, %d, %d, %d, %d et %d\n", cpt_elt,
           (*cel_val_fac)[cpt_elt*6    ], (*cel_val_fac)[cpt_elt*6 + 1],
           (*cel_val_fac)[cpt_elt*6 + 2], (*cel_val_fac)[cpt_elt*6 + 3],
           (*cel_val_fac)[cpt_elt*6 + 4], (*cel_val_fac)[cpt_elt*6 + 5]);

#endif

  /* Hirarchie des cellules */

  BFT_MALLOC(cel_val_hierarch_lu, *nbr_cel * 2, ecs_int_32_t);

  cpt_elt = 0 ;

  if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {

    bft_file_read(cel_val_hierarch_lu, sizeof(ecs_int_32_t),
                  *nbr_cel * 2, fic_maillage) ;

  }
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {

    for (cpt_elt = 0 ;
         cpt_elt < *nbr_cel ;
         cpt_elt++) {

      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;

      retour = sscanf(chaine, "%*d %d %d",
                      &(cel_val_hierarch_lu[cpt_elt * 2    ]),
                      &(cel_val_hierarch_lu[cpt_elt * 2 + 1]));
      if (retour != 2)
        bft_error(__FILE__, __LINE__, errno,
                  _("Error reading line %d of file \"%s\"."),
                  *num_ligne, bft_file_get_name(fic_maillage));

    }

  }

  if (sizeof(ecs_int_t) == sizeof(ecs_int_32_t)) {
    (*cel_val_hierarch) = (ecs_int_t *) cel_val_hierarch_lu ;
    cel_val_hierarch_lu = NULL ;
  }
  else {
    ecs_int_t ind ;
    BFT_MALLOC((*cel_val_hierarch), *nbr_cel * 2, ecs_int_t);
    for (ind = 0 ; ind < *nbr_cel * 2 ; ind++)
      (*cel_val_hierarch)[ind] = (ecs_int_t) cel_val_hierarch_lu[ind] ;
    BFT_FREE(cel_val_hierarch_lu) ;
  }

#if 0 && defined(DEBUG) && !defined(NDEBUG)

  for (cpt_elt = 0 ; cpt_elt < *nbr_cel ; cpt_elt++)
    printf("cel %d : hierarch %d %d\n", cpt_elt,
           (*cel_val_hierarch)[cpt_elt * 2    ],
           (*cel_val_hierarch)[cpt_elt * 2 + 1]);

#endif

}


/*----------------------------------------------------------------------------*
 *                      Lecture des liens avec la CAO (rfrences)
 *----------------------------------------------------------------------------*/

static void ecs_loc_pre_hex__lit_references
(
 bft_file_t     *const   fic_maillage    , /* --> Descripteur fichier  lire  */
 ecs_int_t      *const   num_ligne       , /* <-> Compteur des lignes lues    */
 ecs_int_t      *const   nbr_fac_cao     , /* <-- Nbr. faces dans la CAO      */
 ecs_int_t    * *const   nbr_fac_fac_cao , /* <-- Nbr. faces par face CAO     */
 ecs_int_t  * * *const   lis_fac_fac_cao   /* <-- Liste faces par face CAO    */
)
{

  ecs_int_t     retour ;               /* Retour fonctions `scan' pour test   */
  char          chaine[ECS_LOC_LNG_MAX_CHAINE_HEX] ;           /* Ligne lue   */

  ecs_int_t ind ;
  ecs_int_t ind_sub ;

  /* Variables Hex lues */

  ecs_int_32_t   nb_domain_vertex ;
  ecs_int_32_t   nb_domain_edge ;
  ecs_int_32_t   nb_domain_face ;
  ecs_int_32_t   nb_periodic ;

  ecs_int_32_t  *nb_edge ;
  ecs_int_32_t  *nb_face ;
  ecs_int_32_t  *nb_edgevertex ;
  ecs_int_32_t  *nb_facevertex ;
  ecs_int_32_t  *vertex ;
  ecs_int_32_t  *periodic ;

  ecs_int_32_t **edge ;
  ecs_int_32_t **edgevertex ;
  ecs_int_32_t **face ;
  ecs_int_32_t **facevertex ;

  /* Type du fichier de maillage */

  bft_file_type_t  type_fic_maillage ;


  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  /* Type de maillage */

  type_fic_maillage = bft_file_get_type(fic_maillage) ;


  /* Dimensions */
  /*============*/


  /* Nombre de coins, de courbes, et de faces CAO */

  if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {

    bft_file_read(&nb_domain_vertex, sizeof(ecs_int_32_t), 1, fic_maillage) ;
    bft_file_read(&nb_domain_edge,   sizeof(ecs_int_32_t), 1, fic_maillage) ;
    bft_file_read(&nb_domain_face,   sizeof(ecs_int_32_t), 1, fic_maillage) ;

  }
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {

    bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                  fic_maillage, num_ligne) ;
    retour = sscanf(chaine, " %d %d %d",
                    &nb_domain_vertex, &nb_domain_edge, &nb_domain_face);
    if (retour != 3)
      bft_error(__FILE__, __LINE__, errno,
                _("Error reading line %d of file \"%s\"."),
                *num_ligne, bft_file_get_name(fic_maillage));

  }

#if 0 && defined(DEBUG) && !defined(NDEBUG)

  printf("nb_domain_vertex : %d, nb_domain_edge : %d, nb_domain_face : %d\n",
         nb_domain_vertex, nb_domain_edge, nb_domain_face);

#endif

  BFT_MALLOC(nb_edge, nb_domain_edge, ecs_int_32_t) ;
  BFT_MALLOC(nb_face, nb_domain_face, ecs_int_32_t) ;
  BFT_MALLOC(nb_edgevertex, nb_domain_edge, ecs_int_32_t) ;
  BFT_MALLOC(nb_facevertex, nb_domain_face, ecs_int_32_t) ;
  BFT_MALLOC(vertex, nb_domain_vertex, ecs_int_32_t) ;

  BFT_MALLOC(edge, nb_domain_edge, ecs_int_32_t *) ;
  BFT_MALLOC(edgevertex, nb_domain_edge, ecs_int_32_t *) ;
  BFT_MALLOC(face, nb_domain_face, ecs_int_32_t *) ;
  BFT_MALLOC(facevertex, nb_domain_face, ecs_int_32_t *) ;

  if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {

    if (nb_domain_edge > 0)
      bft_file_read(nb_edge, sizeof(ecs_int_32_t),
                    nb_domain_edge, fic_maillage) ;

    if (nb_domain_face > 0)
      bft_file_read(nb_face, sizeof(ecs_int_32_t),
                    nb_domain_face, fic_maillage) ;

    if (nb_domain_edge > 0)
      bft_file_read(nb_edgevertex, sizeof(ecs_int_32_t),
                    nb_domain_edge, fic_maillage) ;

    if (nb_domain_face > 0)
      bft_file_read(nb_facevertex, sizeof(ecs_int_32_t),
                    nb_domain_face, fic_maillage) ;

    if (nb_domain_vertex > 0)
      bft_file_read(vertex, sizeof(ecs_int_32_t),
                    nb_domain_vertex, fic_maillage) ;

  }
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {

    bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                  fic_maillage, num_ligne) ;
    for (ind = 0 ; ind < nb_domain_vertex ; ind++) {
      vertex[ind] = ecs_loc_pre_hex__scan_d_ligne(fic_maillage,
                                                  num_ligne,
                                                  chaine,
                                                  ECS_LOC_LNG_MAX_CHAINE_HEX);
    }

  }

  for (ind = 0; ind < nb_domain_edge; ind++) {

    /* Si le fichier est binaire, nb_edge a dj t lu */
    if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {
      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;
      retour = sscanf(chaine, " %d", &(nb_edge[ind]));
      if (retour != 1)
        bft_error(__FILE__, __LINE__, errno,
                  _("Error reading line %d of file \"%s\"."),
                  *num_ligne, bft_file_get_name(fic_maillage));
    }

    BFT_MALLOC(edge[ind], nb_edge[ind], ecs_int_32_t);

    if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {
      if (nb_edge[ind] > 0)
        bft_file_read(edge[ind], sizeof(ecs_int_32_t),
                      nb_edge[ind], fic_maillage) ;

    }
    else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {
      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;
      for (ind_sub = 0 ; ind_sub < nb_edge[ind] ; ind_sub++) {
        edge[ind][ind_sub]
          = ecs_loc_pre_hex__scan_d_ligne(fic_maillage,
                                          num_ligne,
                                          chaine,
                                          ECS_LOC_LNG_MAX_CHAINE_HEX);
      }
    }

    /* Si le fichier est binaire, nb_edgevertex a dj t lu */
    if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {
      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;
      retour = sscanf(chaine, " %d", &(nb_edgevertex[ind]));
      if (retour != 1)
        bft_error(__FILE__, __LINE__, errno,
                  _("Error reading line %d of file \"%s\"."),
                  *num_ligne, bft_file_get_name(fic_maillage));
    }

    BFT_MALLOC(edgevertex[ind], nb_edgevertex[ind], ecs_int_32_t);

    if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {
      if (nb_edgevertex[ind] > 0)
        bft_file_read(edgevertex[ind], sizeof(ecs_int_32_t),
                      nb_edgevertex[ind], fic_maillage) ;
    }
    else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {
      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;
      for (ind_sub = 0 ; ind_sub < nb_edgevertex[ind] ; ind_sub++) {
        edgevertex[ind][ind_sub]
          = ecs_loc_pre_hex__scan_d_ligne(fic_maillage,
                                          num_ligne,
                                          chaine,
                                          ECS_LOC_LNG_MAX_CHAINE_HEX);
      }
    }

  }

  for (ind = 0; ind < nb_domain_face; ind++) {

    /* Si le fichier est binaire, nb_face a dj t lu */
    if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {
      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;
      retour = sscanf(chaine, " %d", &(nb_face[ind]));
      if (retour != 1)
        bft_error(__FILE__, __LINE__, errno,
                  _("Error reading line %d of file \"%s\"."),
                  *num_ligne, bft_file_get_name(fic_maillage));
    }

    BFT_MALLOC(face[ind], nb_face[ind], ecs_int_32_t);

    if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {
      if (nb_face[ind] > 0)
        bft_file_read(face[ind], sizeof(ecs_int_32_t),
                      nb_face[ind], fic_maillage) ;
    }
    else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {
      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;
      for (ind_sub = 0 ; ind_sub < nb_face[ind] ; ind_sub++) {
        face[ind][ind_sub]
          = ecs_loc_pre_hex__scan_d_ligne(fic_maillage,
                                          num_ligne,
                                          chaine,
                                          ECS_LOC_LNG_MAX_CHAINE_HEX);
      }
    }

    /* Si le fichier est binaire, nb_facevertex a dj t lu */
    if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {
      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;
      retour = sscanf(chaine, " %d", &(nb_facevertex[ind]));
      if (retour != 1)
        bft_error(__FILE__, __LINE__, errno,
                  _("Error reading line %d of file \"%s\"."),
                  *num_ligne, bft_file_get_name(fic_maillage));
    }

    BFT_MALLOC(facevertex[ind], nb_facevertex[ind], ecs_int_32_t);

    if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {
      if (nb_facevertex[ind] > 0)
        bft_file_read(facevertex[ind], sizeof(ecs_int_32_t),
                      nb_facevertex[ind], fic_maillage) ;
    }
    else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {
      bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                    fic_maillage, num_ligne) ;
      for (ind_sub = 0 ; ind_sub < nb_facevertex[ind] ; ind_sub++) {
        facevertex[ind][ind_sub]
          = ecs_loc_pre_hex__scan_d_ligne(fic_maillage,
                                          num_ligne,
                                          chaine,
                                          ECS_LOC_LNG_MAX_CHAINE_HEX);
      }
    }

  }


  /* Priodicit */

  if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {

    bft_file_read(&nb_periodic, sizeof(ecs_int_32_t), 1, fic_maillage) ;

  }
  else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {

    bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                  fic_maillage, num_ligne) ;
    retour = sscanf(chaine, " %d", &nb_periodic);
    if (retour != 1)
      bft_error(__FILE__, __LINE__, errno,
                _("Error reading line %d of file \"%s\"."),
                *num_ligne, bft_file_get_name(fic_maillage));
  }

  BFT_MALLOC(periodic, nb_periodic * 2, ecs_int_32_t);

  if (nb_periodic > 0) {

    if (type_fic_maillage == BFT_FILE_TYPE_BINARY) {
      bft_file_read(periodic, sizeof(ecs_int_32_t),
                    nb_periodic * 2, fic_maillage) ;
    }
    else if (type_fic_maillage == BFT_FILE_TYPE_TEXT) {
      for (ind_sub = 0 ; ind_sub < nb_facevertex[ind] ; ind_sub++) {
        bft_file_gets(chaine, ECS_LOC_LNG_MAX_CHAINE_HEX,
                      fic_maillage, num_ligne) ;
        retour = sscanf(chaine, "%d %d",
                        &(periodic[ind * 2    ]),
                        &(periodic[ind * 2 + 1]));
        if (retour != 2)
          bft_error(__FILE__, __LINE__, errno,
                    _("Error reading line %d of file \"%s\"."),
                    *num_ligne, bft_file_get_name(fic_maillage));
      }
    }

  }

#if 0 && defined(DEBUG) && !defined(NDEBUG)

  printf("nb_edge :\n");
  for (ind = 0; ind < nb_domain_edge; ind++) {
    printf("%d\n", nb_edge[ind]);
    for (ind_sub = 0 ; ind_sub < nb_edge[ind] ; ind_sub++)
      printf("%d", edge[ind][ind_sub]);
    printf("\n");
  }

  printf("nb_edgevertex :\n");
  for (ind = 0; ind < nb_domain_edge; ind++) {
    printf("%d\n", nb_edgevertex[ind]);
    for (ind_sub = 0 ; ind_sub < nb_edgevertex[ind] ; ind_sub++)
      printf("%d", edgevertex[ind][ind_sub]);
    printf("\n");
  }

  printf("nb_face :\n");
  for (ind = 0; ind < nb_domain_face; ind++) {
    printf("%d\n", nb_face[ind]);
    for (ind_sub = 0 ; ind_sub < nb_face[ind] ; ind_sub++)
      printf("%d", face[ind][ind_sub]);
    printf("\n");
  }

  printf("nb_facevertex :\n");
  for (ind = 0; ind < nb_domain_face; ind++) {
    printf("%d\n", nb_facevertex[ind]);
    for (ind_sub = 0 ; ind_sub < nb_facevertex[ind] ; ind_sub++)
      printf("%d", facevertex[ind][ind_sub]);
    printf("\n");
  }

  printf("\n\nvertex :");
  for (ind = 0 ; ind < nb_domain_vertex ; ind++)
    printf (" %d", vertex[ind]);

  printf("\n\nperiodic : %d\n", nb_periodic);
  for (ind = 0 ; ind < nb_periodic ; ind++)
    printf ("%d %d\n", periodic[ind * 2], periodic[ind * 2 + 1]);

#endif

  /*
    Libration des tableaux inutiliss (on ne conserve que les liens
    entre les faces du maillage et les faces CAO ; la priodicit lue pourra
    aussi tre conserve lorsqu'elle sera implante dans le module enveloppe).
  */

  BFT_FREE(vertex) ;

  for (ind = 0; ind < nb_domain_edge; ind++)
    BFT_FREE(edge[ind]);
  BFT_FREE(edge);
  BFT_FREE(nb_edge) ;

  for (ind = 0; ind < nb_domain_edge; ind++)
    BFT_FREE(edgevertex[ind]);
  BFT_FREE(edgevertex);
  BFT_FREE(nb_edgevertex) ;

  for (ind = 0; ind < nb_domain_face; ind++)
    BFT_FREE(facevertex[ind]);
  BFT_FREE(facevertex);
  BFT_FREE(nb_facevertex) ;

  BFT_FREE(periodic);


  /* On renvoie les liens sur les faces */

  if (sizeof (ecs_int_32_t) == sizeof (ecs_int_t)) {
    *nbr_fac_cao = nb_domain_face ;
    *nbr_fac_fac_cao = nb_face ;
    *lis_fac_fac_cao = face ;
  }
  else {
    *nbr_fac_cao = (ecs_int_t) nb_domain_face ;
    BFT_MALLOC((*nbr_fac_fac_cao), nb_domain_face, ecs_int_t) ;
    BFT_MALLOC((*lis_fac_fac_cao), nb_domain_face, ecs_int_t *) ;
    for (ind = 0; ind < nb_domain_face; ind++) {
      (*nbr_fac_fac_cao)[ind] = nb_face[ind] ;
      BFT_MALLOC((*lis_fac_fac_cao)[ind], nb_face[ind], ecs_int_t) ;
      for (ind_sub = 0 ; ind_sub < nb_face[ind] ; ind_sub++)
        (*lis_fac_fac_cao)[ind][ind_sub] =  face[ind][ind_sub] ;
      BFT_FREE(face[ind]) ;
    }
    BFT_FREE(face);
    BFT_FREE(nb_face);
  }


}


/*----------------------------------------------------------------------------*
 *    Affectation des rfrences des faces  la CAO aux faces du maillage
 *----------------------------------------------------------------------------*/

static ecs_int_t * ecs_loc_pre_hex__prep_reference
(
 ecs_int_t       const nbr_fac           , /* --> Nombre de faces             */
 ecs_int_t      *const fac_val_hierarch  , /* --> Hirarchie des faces        */
 ecs_int_t      *const nbr_fac_cao       , /* <-- Nbr. faces dans la CAO      */
 ecs_int_t    * *const nbr_fac_fac_cao   , /* <-- Nbr. faces par face CAO     */
 ecs_int_t  * * *const lis_fac_fac_cao     /* <-- Liste faces par face CAO    */
)
{

  ecs_int_t   ind ;
  ecs_int_t   ind_sub ;
  ecs_int_t   ind_fac ;
  ecs_int_t   cpt_mod ;

  ecs_int_t  *reference ;

  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  /*
    Il semble qu'une face ne puisse tre associe qu' une surface CAO,
    mais dans la structure hirarchique, seule la face la plus
    proche de la racine est donne dans la liste des associations ;
    toutes ses sous-faces ont alors la mme association par dfaut.
    ( vrifier ; si cela n'est pas le cas, on devra soit donner une
    priorit  une face CAO, soit utiliser la notion de groupe).
  */

  /*
    Initialisation du tableau des rfrences
    et dsallocation des listes d'associations
  */

  BFT_MALLOC(reference, nbr_fac, ecs_int_t) ;
  for (ind = 0 ; ind < nbr_fac ; ind++)
    reference[ind] = 0 ;

  for (ind = 0 ; ind < *nbr_fac_cao ; ind++) {

    for (ind_sub = 0 ; ind_sub < (*nbr_fac_fac_cao)[ind] ; ind_sub++) {
      ind_fac = (*lis_fac_fac_cao)[ind][ind_sub] ;
      if (reference[ind_fac] == 0)
        reference[ind_fac] = ind + 1 ;
      else
        bft_error(__FILE__, __LINE__, 0,
                  "La face %d du maillage IggHexa est attache  au moins deux\n"
                  "faces de la CAO (%d et %d).\n"
                  "Ce cas n'est pas prvu pour l'instant.",
                  ind_fac, reference[ind_fac], ind) ;

    }

    BFT_FREE((*lis_fac_fac_cao)[ind]) ;

  }

  BFT_FREE(*lis_fac_fac_cao) ;
  BFT_FREE(*nbr_fac_fac_cao) ;

  *nbr_fac_cao = 0 ;


  /* Propagation rcursive aux faces issues de faces rfrences */

  do {

    cpt_mod = 0 ;

    for (ind = 0 ; ind < nbr_fac ; ind++) {

      if (reference[ind] > 0) {

        for (ind_sub = 0 ; ind_sub < 2 ; ind_sub++) {

          ind_fac = fac_val_hierarch[ind * 2  + ind_sub] ;
          if (ind_fac > -1 && reference[ind_fac] == 0) {
            reference[ind_fac] = reference[ind] ;
            cpt_mod++;
          }

        }

      }

    }

  } while (cpt_mod > 0) ;


  return reference ;

}


/*----------------------------------------------------------------------------*
 *    Suppression des cellules hirarchiques et compactage associ
 *----------------------------------------------------------------------------*/

static void ecs_loc_pre_hex__compacte
(
 ecs_int_t      *const nbr_som           , /* --> Nombre de sommets           */
 ecs_int_t      *const nbr_are           , /* --> Nombre d'artes             */
 ecs_int_t      *const nbr_fac           , /* --> Nombre de faces             */
 ecs_int_t      *const nbr_cel           , /* <-> Nombre de cellules          */
 ecs_real_t   * *const som_val_coord     , /* --> Coordonnes des sommets     */
 ecs_int_t    * *const are_val_som       , /* --> Connectivit sommets/artes */
 ecs_int_t    * *const fac_val_are       , /* --> Connectivit artes/faces   */
 ecs_int_t    * *const cel_val_fac       , /* --> Connectivit faces/cellules */
 ecs_int_t    * *const fac_val_hierarch  , /* --> Hirarchie des faces        */
 ecs_int_t    * *const cel_val_hierarch  , /* --> Hirarchie des cellules     */
 ecs_int_t    * *const fac_val_reference   /* --> Rfrences des faces        */
)
{

  ecs_int_t   ind ;
  ecs_int_t   ind_loc ;
  ecs_int_t   cpt_elt ;
  ecs_int_t   cpt_mod ;

  ecs_int_t  *masque ;
  ecs_int_t  *renum ;

  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  /* Cellules */
  /*==========*/

  /* On ne conservera que les cellules qui n'ont pas de cellules filles */

  BFT_MALLOC(masque, *nbr_cel, ecs_int_t) ;

  for (ind = 0 ; ind < *nbr_cel ; ind++) {
    if ((*cel_val_hierarch)[ind*2] != -1)
      masque[ind] = 0 ;
    else
      masque[ind] = 1 ;
  }

  BFT_FREE(*cel_val_hierarch) ;

  /* Compactage de la dfinition des cellules */

  cpt_elt = 0 ;

  for (ind = 0 ; ind < *nbr_cel ; ind++) {

    if (masque[ind] == 1) {

      for (ind_loc = 0 ; ind_loc < 6 ; ind_loc++)
        (*cel_val_fac)[cpt_elt * 6 + ind_loc]
          = (*cel_val_fac)[ind * 6 + ind_loc] ;

      cpt_elt++ ;

    }

  }

  BFT_FREE(masque) ;

#if 0 && defined(DEBUG) && !defined(NDEBUG)
  printf("Nombre de cellules initiales (avec hirarchie) : %d\n"
         "Nombre de cellules finales (sans hirarchie) : %d\n",
         *nbr_cel, cpt_elt);
#endif

  *nbr_cel = cpt_elt ;

  BFT_REALLOC(*cel_val_fac, (*nbr_cel) * 6, ecs_int_t) ;


  /* Faces */
  /*=======*/

  /*
    Premier compactage : on ne conserve que les faces rfrences
    par les cellules compactes
  */

  BFT_MALLOC(masque, *nbr_fac, ecs_int_t) ;
  BFT_MALLOC(renum, *nbr_fac, ecs_int_t) ;

  for (ind = 0 ; ind < *nbr_fac ; ind++)
    masque[ind] = 0 ;

  for (ind = 0 ; ind < *nbr_cel ; ind++) {
    for (ind_loc = 0 ; ind_loc < 6 ; ind_loc++)
      masque[(*cel_val_fac)[ind * 6 + ind_loc]] = 1 ;
  }

  do {

    cpt_mod = 0 ;

    for (ind = 0 ; ind < *nbr_fac ; ind++) {

      for (ind_loc = 0 ; ind_loc < 2 ; ind_loc++) {

        if (   (masque[ind] == 1)
            && ((*fac_val_hierarch)[ind * 2 + ind_loc] != -1)
               && (masque[(*fac_val_hierarch)[ind*2 + ind_loc]]) == 0) {
          (*fac_val_hierarch)[ind*2 + ind_loc] = -1 ;
          cpt_mod += 1 ;
        }

      }

    }

  } while (cpt_mod != 0) ;


  cpt_elt = 0 ;

  for (ind = 0 ; ind < *nbr_fac ; ind++) {

    if (masque[ind] == 1) {

      for (ind_loc = 0 ; ind_loc < 4 ; ind_loc++)
        (*fac_val_are)[cpt_elt * 4 + ind_loc]
          = (*fac_val_are)[ind * 4 + ind_loc] ;

      (*fac_val_reference)[cpt_elt] = (*fac_val_reference)[ind] ;

      renum[ind] = cpt_elt ;

      cpt_elt++ ;

    }
    else {

      renum[ind] = -1 ;

    }

  }

  cpt_elt = 0 ;

  /* Compactage avec renumrotation de la hirarchie */

  for (ind = 0 ; ind < *nbr_fac ; ind++) {

    if (masque[ind] == 1) {

      for (ind_loc = 0 ; ind_loc < 2 ; ind_loc++) {

        if ((*fac_val_hierarch)[ind * 2 + ind_loc] != -1)
          (*fac_val_hierarch)[cpt_elt * 2 + ind_loc]
            = renum[(*fac_val_hierarch)[ind * 2 + ind_loc]] ;
        else
          (*fac_val_hierarch)[cpt_elt * 2 + ind_loc] = -1 ;

      }

      cpt_elt++ ;

    }

  }

  /* Renumrotation dans la connectivit cellules / faces */

  for (ind = 0 ; ind < *nbr_cel * 6 ; ind++) {
    (*cel_val_fac)[ind] = renum[(*cel_val_fac)[ind]] ;
    assert ((*cel_val_fac)[ind] > -1) ;
  }


  BFT_FREE(renum) ;
  BFT_FREE(masque) ;

#if 0 && defined(DEBUG) && !defined(NDEBUG)
  printf("Nombre de faces initiales (avec hirarchie) : %d\n"
         "Nombre de faces finales (sans hirarchie) : %d\n",
         *nbr_fac, cpt_elt);
#endif

  *nbr_fac = cpt_elt ;

  BFT_REALLOC(*fac_val_are,       (*nbr_fac) * 4, ecs_int_t) ;
  BFT_REALLOC(*fac_val_hierarch,  (*nbr_fac) * 2, ecs_int_t) ;
  BFT_REALLOC(*fac_val_reference, (*nbr_fac),     ecs_int_t) ;


  /* Artes */
  /*========*/

  /*
    Premier compactage : on ne conserve que les artes rfrences
    par les faces compactes.
  */

  BFT_MALLOC(masque, *nbr_are, ecs_int_t) ;
  BFT_MALLOC(renum, *nbr_are, ecs_int_t) ;

  for (ind = 0 ; ind < *nbr_are ; ind++)
    masque[ind] = 0 ;

  for (ind = 0 ; ind < *nbr_fac ; ind++) {
    for (ind_loc = 0 ; ind_loc < 4 ; ind_loc++)
      masque[(*fac_val_are)[ind * 4 + ind_loc]] = 1 ;
  }


  cpt_elt = 0 ;

  for (ind = 0 ; ind < *nbr_are ; ind++) {

    if (masque[ind] == 1) {

      for (ind_loc = 0 ; ind_loc < 2 ; ind_loc++)
        (*are_val_som)[cpt_elt * 2 + ind_loc]
          = (*are_val_som)[ind * 2 + ind_loc] ;

      renum[ind] = cpt_elt ;

      cpt_elt++ ;

    }
    else {

      renum[ind] = -1 ;

    }

  }


  /* Renumrotation dans la connectivit faces / artes */

  for (ind = 0 ; ind < *nbr_fac * 4 ; ind++) {
    (*fac_val_are)[ind] = renum[(*fac_val_are)[ind]] ;
    assert ((*fac_val_are)[ind] > -1) ;
  }

  BFT_FREE(renum) ;
  BFT_FREE(masque) ;

#if 0 && defined(DEBUG) && !defined(NDEBUG)
  printf("Nombre d'artes initiales (avec hirarchie) : %d\n"
         "Nombre d'artes finales (sans hirarchie) : %d\n",
         *nbr_are, cpt_elt);
#endif

  *nbr_are = cpt_elt ;

  BFT_REALLOC(*are_val_som,       (*nbr_are) * 2, ecs_int_t) ;


  /* Sommets */
  /*=========*/

  /*
    Premier compactage : on ne conserve que les sommets rfrencs
    par les artes compactes.
  */

  BFT_MALLOC(masque, *nbr_som, ecs_int_t) ;
  BFT_MALLOC(renum, *nbr_som, ecs_int_t) ;

  for (ind = 0 ; ind < *nbr_som ; ind++)
    masque[ind] = 0 ;

  for (ind = 0 ; ind < *nbr_are ; ind++) {
    for (ind_loc = 0 ; ind_loc < 2 ; ind_loc++)
      masque[(*are_val_som)[ind * 2 + ind_loc]] = 1 ;
  }


  cpt_elt = 0 ;

  for (ind = 0 ; ind < *nbr_som ; ind++) {

    if (masque[ind] == 1) {

      for (ind_loc = 0 ; ind_loc < 3 ; ind_loc++)
        (*som_val_coord)[cpt_elt * 3 + ind_loc]
          = (*som_val_coord)[ind * 3 + ind_loc] ;

      renum[ind] = cpt_elt ;

      cpt_elt++ ;

    }
    else {

      renum[ind] = -1 ;

    }

  }

  /* Renumrotation dans la connectivit artes / sommets */

  for (ind = 0 ; ind < *nbr_are * 2 ; ind++) {
    (*are_val_som)[ind] = renum[(*are_val_som)[ind]] ;
    assert ((*are_val_som)[ind] > -1) ;
  }

  BFT_FREE(renum) ;
  BFT_FREE(masque) ;

#if 0 && defined(DEBUG) && !defined(NDEBUG)
  printf("Nombre de sommets initiaux (avec hirarchie) : %d\n"
         "Nombre de sommets finaux (sans hirarchie) : %d\n",
         *nbr_som, cpt_elt);
#endif

  *nbr_som = cpt_elt ;

  BFT_REALLOC(*som_val_coord, (*nbr_som) * 3, ecs_real_t) ;


}


/*----------------------------------------------------------------------------*
 *    Construction de la connectivit nodale
 *----------------------------------------------------------------------------*/

static void ecs_loc_pre_hex__connect_nod
(
 ecs_int_t      *const nbr_are           , /* <-> Nombre d'artes             */
 ecs_int_t      *const nbr_fac           , /* <-> Nombre de faces             */
 ecs_int_t             nbr_cel           , /* --> Nombre de cellules          */
 ecs_real_t     *const som_val_coord     , /* --> Coordonnes des sommets     */
 ecs_int_t    * *const are_val_som       , /* <-> Connectivit sommets/artes */
 ecs_int_t    * *const fac_val_are       , /* <-> Connectivit artes/faces   */
 ecs_int_t    * *const fac_val_som       , /* <-- Connectivit sommets/faces  */
 ecs_int_t    * *const cel_val_fac       , /* <-> Connectivit faces/cellules */
 ecs_int_t    * *const cel_val_som       , /* <-- Connectivit sommets/cel.   */
 ecs_int_t    * *const fac_val_hierarch  , /* <-> Hirarchie des faces        */
 ecs_int_t    * *const fac_val_reference   /* <-> Rfrences des faces        */
)
{

  ecs_int_t     icoo ;
  ecs_int_t     ind ;
  ecs_int_t     ind_are ;
  ecs_int_t     ind_cel ;
  ecs_int_t     ind_fac ;
  ecs_int_t     ind_fac_loc ;
  ecs_int_t     ind_som_loc ;
  ecs_int_t     ind_som ;
  ecs_int_t     ind_som_1 ;
  ecs_int_t     ind_som_2 ;
  ecs_int_t     pos_deb ;
  ecs_int_t     ind_loc ;
  ecs_int_t     ipass ;
  ecs_int_t     parcours ;
  ecs_int_t     cpt_elt ;
  ecs_int_t     cpt_mod ;

  ecs_point_t   coord_cdg_cel ;
  ecs_point_t   coord_cdg_fac ;
  ecs_point_t   vect_1 ;
  ecs_point_t   vect_2 ;
  ecs_point_t   prod_vect;
  ecs_point_t   vect_comp ;

  ecs_int_t    *masque ;
  ecs_int_t    *renum ;

  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  /* Premier traitement des faces */
  /*==============================*/


  /* Construction de la connectivit nodale */

  BFT_MALLOC(*fac_val_som, (*nbr_fac) * 4, ecs_int_t) ;

  for (ind_fac = 0 ; ind_fac < *nbr_fac ; ind_fac++) {

    pos_deb = ind_fac * 4 ;

    (*fac_val_som)[pos_deb    ]
      = (*are_val_som)[((*fac_val_are)[pos_deb]) * 2    ] ;

    (*fac_val_som)[pos_deb + 1]
      = (*are_val_som)[((*fac_val_are)[pos_deb]) * 2 + 1] ;

    /*
      On a les sommets 1 et 2, mais on ne sait pas encore quel sens de
      parcours adopter pour passer au sommet 3.
    */

    ind_som_1 = (*are_val_som)[((*fac_val_are)[pos_deb + 1]) * 2    ] ;
    ind_som_2 = (*are_val_som)[((*fac_val_are)[pos_deb + 1]) * 2 + 1] ;

    if (   ind_som_1 == (*fac_val_som)[pos_deb + 1]
        || ind_som_2 == (*fac_val_som)[pos_deb + 1])
      parcours = 1 ;
    else
      parcours = -1 ;

    for (ind_loc = 1 ; ind_loc < 3 ; ind_loc++) {

      ind_are = (*fac_val_are)[pos_deb + ((4 + (parcours * ind_loc)) % 4)] ;

      ind_som_1 = (*are_val_som)[ind_are * 2    ] ;
      ind_som_2 = (*are_val_som)[ind_are * 2 + 1] ;

      if (ind_som_1 == (*fac_val_som)[pos_deb + ind_loc])
        (*fac_val_som)[pos_deb + ind_loc + 1] = ind_som_2 ;
      else if (ind_som_2 == (*fac_val_som)[pos_deb + ind_loc])
        (*fac_val_som)[pos_deb + ind_loc + 1] = ind_som_1 ;
      else
        assert (   ind_som_1 == (*fac_val_som)[pos_deb + ind_loc]
                || ind_som_2 == (*fac_val_som)[pos_deb + ind_loc]);

    }

#if 0 && defined(DEBUG) && !defined(NDEBUG)

    printf("Face : %d\n", ind_fac);
    for (ind_som_loc = 0 ; ind_som_loc < 4 ; ind_som_loc++) {
      ind_som = (*fac_val_som)[ind_fac * 4 + ind_som_loc] ;
      printf ("sommet %d : %f %f %f\n", ind_som,
              som_val_coord[ind_som * 3    ],
              som_val_coord[ind_som * 3 + 1],
              som_val_coord[ind_som * 3 + 2]);
    }

#endif

  }


  /* On n'a plus besoin de la connectivit faces / artes */

  BFT_FREE(*fac_val_are) ;


  /* Artes */
  /*========*/

  /* On n'a plus besoin des artes */

  *nbr_are = 0 ;

  BFT_FREE(*are_val_som) ;


  /* Cellules */
  /*==========*/

  /* Connectivit cellules / sommets */

  BFT_MALLOC(*cel_val_som, nbr_cel * 8, ecs_int_t) ;


  for (ind_cel = 0 ; ind_cel < nbr_cel ; ind_cel++) {

    /*
      Centre de gravit des sommets des faces (proche du centre de l'hxadre)
    */

    coord_cdg_cel[0] = 0.0 ; coord_cdg_cel[1] = 0.0 ; coord_cdg_cel[2] = 0.0 ;

    for (ind_fac_loc = 0 ; ind_fac_loc < 6 ; ind_fac_loc++) {

      ind_fac = (*cel_val_fac)[ind_cel * 6 + ind_fac_loc] ;

      for (ind_som_loc = 0 ; ind_som_loc < 4 ; ind_som_loc++) {

        ind_som = (*fac_val_som)[ind_fac * 4 + ind_som_loc] ;

        for (icoo = 0 ; icoo < ECS_DIM_3 ; icoo++)
          coord_cdg_cel[icoo] += som_val_coord[ind_som * ECS_DIM_3 + icoo] ;

      }

    }

    for (icoo = 0 ; icoo < ECS_DIM_3 ; icoo++)
      coord_cdg_cel[icoo] /= 24.0 ;


    /* Centre de gravit et normale de la premire face */

    coord_cdg_fac[0] = 0.0 ; coord_cdg_fac[1] = 0.0 ; coord_cdg_fac[2] = 0.0 ;

    ind_fac = (*cel_val_fac)[ind_cel * 6] ;

    for (ind_som_loc = 0 ; ind_som_loc < 4 ; ind_som_loc++) {

      ind_som = (*fac_val_som)[ind_fac * 4 + ind_som_loc] ;

      for (icoo = 0 ; icoo < ECS_DIM_3 ; icoo++)
        coord_cdg_fac[icoo] += som_val_coord[ind_som * ECS_DIM_3 + icoo] ;

    }

    for (icoo = 0 ; icoo < ECS_DIM_3 ; icoo++)
      coord_cdg_fac[icoo] /= 4.0 ;

    for (icoo = 0 ; icoo < ECS_DIM_3 ; icoo++) {
      vect_1[icoo]
        =   som_val_coord[(*fac_val_som)[ind_fac * 4 + 1] * ECS_DIM_3 + icoo]
          - som_val_coord[(*fac_val_som)[ind_fac * 4    ] * ECS_DIM_3 + icoo] ;
      vect_2[icoo]
        =   som_val_coord[(*fac_val_som)[ind_fac * 4 + 3] * ECS_DIM_3 + icoo]
          - som_val_coord[(*fac_val_som)[ind_fac * 4    ] * ECS_DIM_3 + icoo] ;
    }

    ECS_LOC_PRODUIT_VECTORIEL(prod_vect, vect_1, vect_2) ;

    for (icoo = 0 ; icoo < ECS_DIM_3 ; icoo++)
      vect_comp[icoo] = coord_cdg_fac[icoo] - coord_cdg_cel[icoo] ;


    /*
      Si la normale pointe vers l'extrieur, on prend les sommets en sens
      inverse (numrotation face "infrieure" de l'hexadre).

      *     8 x-------x 7
      *      /|      /|
      *     / |     / |
      *  5 x-------x6 |
      *    | 4x----|--x 3
      *    | /     | /
      *    |/      |/
      *  1 x-------x 2

    */

    if (ECS_LOC_PRODUIT_SCALAIRE(vect_comp, prod_vect) > 0.0) {

      /* normale extrieure */

      for (ind_som_loc = 0 ; ind_som_loc < 4 ; ind_som_loc++) {
        ind_som = (*fac_val_som)[ind_fac * 4 + ind_som_loc] ;
        (*cel_val_som)[ind_cel * 8 + 3 - ind_som_loc] = ind_som ;
      }

    }
    else {

      /* normale intrieure */

      for (ind_som_loc = 0 ; ind_som_loc < 4 ; ind_som_loc++) {
        ind_som = (*fac_val_som)[ind_fac * 4 + ind_som_loc] ;
        (*cel_val_som)[ind_cel * 8 + ind_som_loc] = ind_som ;
      }

    }

    /*
      On a 4 sommets sur 8 ; on cherche maintenant la face partageant
      les sommets 1 et 2, afin de connatre les sommets 5 et 6,
      puis la face partageant les sommets 3 et 4, afin de connatre
      les sommets 7 et 8.
    */

    for (ipass = 0 ; ipass < 2 ; ipass++) {

      ind_som_1 = (*cel_val_som)[ind_cel * 8     + (ipass * 2)] ;
      ind_som_2 = (*cel_val_som)[ind_cel * 8 + 1 + (ipass * 2)] ;

      parcours = 0 ;

      for (ind_fac_loc = 1 ;
           ind_fac_loc < 6 && parcours == 0 ;
           ind_fac_loc++) {

        ind_fac = (*cel_val_fac)[ind_cel * 6 + ind_fac_loc] ;

        for (ind_som_loc = 0 ;
             ind_som_loc < 4 ;
             ind_som_loc++) {

          ind_som = (*fac_val_som)[ind_fac * 4 + ind_som_loc] ;

          if (ind_som == ind_som_1) {

            if ((*fac_val_som)[ind_fac * 4 + ((ind_som_loc + 1) % 4)]
                == ind_som_2) {
              parcours = 1 ;
              break ;
            }

            else if ((*fac_val_som)[ind_fac * 4 + ((ind_som_loc - 1 + 4) % 4)]
                     == ind_som_2) {
              parcours = -1 ;
              break ;
            }
          }

        }

      }

      (*cel_val_som)[ind_cel * 8 + 4 + (ipass * 2)]
        =(*fac_val_som)[ind_fac * 4
                        + ((ind_som_loc + (4 + (3 * parcours))) % 4)] ;

      (*cel_val_som)[ind_cel * 8 + 5 + (ipass * 2)]
        =(*fac_val_som)[ind_fac * 4
                        + ((ind_som_loc + (4 + (2 * parcours))) % 4)] ;

    }

#if 0 && defined(DEBUG) && !defined(NDEBUG)

    printf("Cellule : %d\n", ind_cel);
    for (ind_som_loc = 0 ; ind_som_loc < 8 ; ind_som_loc++) {
      ind_som = (*cel_val_som)[ind_cel * 8 + ind_som_loc] ;
      printf ("sommet %d : %f %f %f\n", ind_som,
              som_val_coord[ind_som * 3    ],
              som_val_coord[ind_som * 3 + 1],
              som_val_coord[ind_som * 3 + 2]);
    }

#endif

  }


  /* On n'a plus besoin de la connectivit cellules / faces */

  BFT_FREE(*cel_val_fac) ;


  /* Second traitement des faces */
  /*==============================*/

  /*
    On n'a plus besoin des faces qui n'interviennent pas dans la hirarchie
    et qui ne portent pas non plus une rfrence
  */

  BFT_MALLOC(masque, *nbr_fac, ecs_int_t) ;
  BFT_MALLOC(renum, *nbr_fac, ecs_int_t) ;

  for (ind = 0 ; ind < *nbr_fac ; ind++)
    if ((*fac_val_reference)[ind] != 0)
      masque[ind] = 1 ;
    else
      masque[ind] = 0 ;

  for (ind = 0 ; ind < *nbr_fac ; ind++) {
    if ((*fac_val_hierarch)[ind * 2] != -1) {
      masque[ind] = 1 ;
      masque[(*fac_val_hierarch)[ind * 2]] = 1 ;
    }
  }


  do {

    cpt_mod = 0 ;

    for (ind = 0 ; ind < *nbr_fac ; ind++) {

      if (masque[ind] == 1 && (*fac_val_hierarch)[ind * 2 + 1] != -1) {

        if (masque[(*fac_val_hierarch)[ind * 2 + 1]] == 0) {

          masque[(*fac_val_hierarch)[ind * 2 + 1]] = 1 ;
          cpt_mod += 1 ;

        }

      }

    }

  } while (cpt_mod != 0) ;


  cpt_elt = 0 ;

  for (ind = 0 ; ind < *nbr_fac ; ind++) {

    if (masque[ind] == 1) {

      for (ind_loc = 0 ; ind_loc < 4 ; ind_loc++)
        (*fac_val_som)[cpt_elt * 4 + ind_loc]
          = (*fac_val_som)[ind * 4 + ind_loc] ;

      for (ind_loc = 0 ; ind_loc < 2 ; ind_loc++)
        (*fac_val_hierarch)[cpt_elt * 2 + ind_loc]
          = (*fac_val_hierarch)[ind * 2 + ind_loc] ;

      (*fac_val_reference)[cpt_elt] = (*fac_val_reference)[ind] ;

      renum[ind] = cpt_elt ;

      cpt_elt++ ;

    }
    else {

      renum[ind] = -1 ;

    }

  }

  for (ind = 0 ; ind < cpt_elt * 2 ; ind++) {
    if ((*fac_val_hierarch)[ind] != -1)
      (*fac_val_hierarch)[ind] = renum[(*fac_val_hierarch)[ind]] ;
  }

  BFT_FREE(masque) ;
  BFT_FREE(renum) ;

#if 0 && defined(DEBUG) && !defined(NDEBUG)
  printf("Nombre de faces initiales (connectivit descendante) : %d\n"
         "Nombre de faces finales (connectivit nodale) : %d\n",
         *nbr_fac, cpt_elt);
#endif

  *nbr_fac = cpt_elt ;

  BFT_REALLOC(*fac_val_reference, (*nbr_fac),     ecs_int_t) ;
  BFT_REALLOC(*fac_val_hierarch,  (*nbr_fac) * 2, ecs_int_t) ;
  BFT_REALLOC(*fac_val_som,       (*nbr_fac) * 4, ecs_int_t) ;

  /*
    Dcalage des connectivits (incrment de 1)
    Les hirarchies sont traites dans une fonction spcifique
  */

  for (ind = 0 ; ind < nbr_cel * 8 ; ind++)
    (*cel_val_som)[ind] += 1 ;

  for (ind = 0 ; ind < (*nbr_fac) * 4 ; ind++)
    (*fac_val_som)[ind] += 1 ;

}


/*----------------------------------------------------------------------------*
 *    Construction de la structure maillage
 *----------------------------------------------------------------------------*/

static ecs_entmail_t ** ecs_loc_pre_hex__cree_entmail
(
 ecs_int_t             nbr_som           , /* --> Nombre de sommets           */
 ecs_int_t             nbr_fac           , /* --> Nombre de faces             */
 ecs_int_t             nbr_cel           , /*  -> Nombre de cellules          */
 ecs_real_t   * *const som_val_coord     , /* --> Coordonnes des sommets     */
 ecs_int_t    * *const are_val_som       , /* <-> Connectivit sommets/artes */
 ecs_int_t    * *const fac_val_som       , /* <-- Connectivit sommets/faces  */
 ecs_int_t    * *const cel_val_som       , /* <-- Connectivit sommets/cel.   */
 ecs_int_t    * *const fac_val_reference   /* <-> Rfrences des faces        */
)
{

  ecs_int_t    ind                  ;
  ecs_int_t    ind_cpt              ;

  ecs_int_t    ient                 ; /* Indice de boucle sur les entits     */
  ecs_int_t    num_coul_max         ; /* Numro de couleur maximal            */
  ecs_int_t    cpt_coul_ent[ECS_ENTMAIL_FIN]; /* Compteur de couleurs         */
  ecs_int_t   *val_coul_ent[ECS_ENTMAIL_FIN]; /* Tableau valeurs des couleurs */
  ecs_size_t  *cpt_elt_coul_ent[ECS_ENTMAIL_FIN];


  /* Stockage avant transfert */
  /*--------------------------*/

  size_t       nbr_elt_ent        [ECS_ENTMAIL_FIN]; /* Nombre d'elems/entite */

  ecs_size_t  *elt_pos_som_ent    [ECS_ENTMAIL_FIN]; /* Positions numeros som */
  ecs_int_t   *elt_val_som_ent    [ECS_ENTMAIL_FIN]; /* Numeros des sommets   */
  ecs_int_t   *elt_val_typ_geo_ent[ECS_ENTMAIL_FIN]; /* Types geometriques    */
  ecs_int_t   *elt_val_color_ent  [ECS_ENTMAIL_FIN]; /* Couleurs              */

  ecs_int_t   *renum ;

  ecs_entmail_t  *vect_entmail_som ;
  ecs_entmail_t **vect_entmail ;


  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  /* Sommets */

  vect_entmail_som = ecs_entmail_pre__cree_som(ECS_DIM_3,
                                               nbr_som,
                                               (*som_val_coord),
                                               NULL) ;

  /* som_val_coord est libr par ecs_entmail_pre__cree_som() */

  *som_val_coord = NULL ;


  /* Initialisations et allocations des tableaux locaux */


  /* Attention au decalage de `1' !!!         */
  /* On n'alloue pas les tableaux locaux pour */
  /* `ECS_ENTMAIL_DEB = ECS_ENTMAIL_SOM'      */

  for (ient = ECS_ENTMAIL_ARE ; ient < ECS_ENTMAIL_FIN ; ient++) {

    nbr_elt_ent        [ient] = 0 ;

    cpt_coul_ent       [ient] = 0    ;
    cpt_elt_coul_ent   [ient] = NULL ;
    val_coul_ent       [ient] = NULL ;

  }


  /* Artes */
  /*--------*/

  nbr_elt_ent[ECS_ENTMAIL_ARE] = 0 ;

  elt_pos_som_ent[ECS_ENTMAIL_ARE] = NULL ;
  elt_val_som_ent[ECS_ENTMAIL_ARE] = NULL ;

  elt_val_typ_geo_ent[ECS_ENTMAIL_ARE] = NULL ;
  elt_val_color_ent[ECS_ENTMAIL_ARE] = NULL ;


  /* Faces */
  /*-------*/

  nbr_elt_ent[ECS_ENTMAIL_FAC] = nbr_fac ;

  BFT_MALLOC(elt_pos_som_ent[ECS_ENTMAIL_FAC], nbr_fac + 1, ecs_size_t) ;

  for (ind = 0 ; ind < nbr_fac + 1; ind++)
    elt_pos_som_ent[ECS_ENTMAIL_FAC][ind] =  (ind * 4 + 1) ;

  elt_val_som_ent[ECS_ENTMAIL_FAC] = (*fac_val_som) ;

  BFT_MALLOC(elt_val_typ_geo_ent[ECS_ENTMAIL_FAC], nbr_fac, ecs_int_t) ;

  for (ind = 0 ; ind < nbr_fac ; ind++)
    elt_val_typ_geo_ent[ECS_ENTMAIL_FAC][ind] = ECS_ELT_TYP_FAC_QUAD ;

  elt_val_color_ent[ECS_ENTMAIL_FAC] = (*fac_val_reference) ;


  /* Cellules */
  /*----------*/

  nbr_elt_ent[ECS_ENTMAIL_CEL] = nbr_cel ;

  BFT_MALLOC(elt_pos_som_ent[ECS_ENTMAIL_CEL], nbr_cel + 1, ecs_size_t) ;

  for (ind = 0 ; ind < nbr_cel + 1; ind++)
    elt_pos_som_ent[ECS_ENTMAIL_CEL][ind] =  (ind * 8 + 1) ;

  elt_val_som_ent[ECS_ENTMAIL_CEL] = (*cel_val_som) ;

  BFT_MALLOC(elt_val_typ_geo_ent[ECS_ENTMAIL_CEL], nbr_cel, ecs_int_t) ;

  for (ind = 0 ; ind < nbr_cel ; ind++)
    elt_val_typ_geo_ent[ECS_ENTMAIL_CEL][ind] = ECS_ELT_TYP_CEL_HEXA ;

  elt_val_color_ent[ECS_ENTMAIL_CEL] = NULL ;


  /* Couleurs (rfrences) des faces */
  /*---------------------------------*/


  num_coul_max = 0 ;

  for (ind = 0 ; ind < nbr_fac ; ind++) {
    if ((*fac_val_reference)[ind] > num_coul_max)
      num_coul_max = (*fac_val_reference)[ind] ;
  }

  BFT_MALLOC(cpt_elt_coul_ent[ECS_ENTMAIL_FAC], num_coul_max + 1, ecs_size_t);

  for (ind = 0 ; ind < num_coul_max + 1; ind++)
    cpt_elt_coul_ent[ECS_ENTMAIL_FAC][ind] = 0 ;

  for (ind = 0 ; ind < nbr_fac ; ind++)
    cpt_elt_coul_ent[ECS_ENTMAIL_FAC][(*fac_val_reference)[ind]] += 1 ;

  /* Compactage des rfrences rellement utilises */

  BFT_MALLOC(val_coul_ent[ECS_ENTMAIL_FAC], num_coul_max + 1, ecs_int_t);
  BFT_MALLOC(renum, num_coul_max + 1, ecs_int_t);

  ind_cpt = 0 ;

  for (ind = 0 ; ind < num_coul_max + 1; ind++) {

    if (cpt_elt_coul_ent[ECS_ENTMAIL_FAC][ind] > 0) {

      val_coul_ent[ECS_ENTMAIL_FAC][ind_cpt] = ind ;
      cpt_elt_coul_ent[ECS_ENTMAIL_FAC][ind_cpt]
        = cpt_elt_coul_ent[ECS_ENTMAIL_FAC][ind] ;

      /* On ajoute 1 pour une numrotation de 1  n et non de 0  n-1 */

      renum[ind] = ind_cpt + 1;

      ind_cpt += 1 ;

    }

  }

  cpt_coul_ent[ECS_ENTMAIL_FAC] = ind_cpt ;

  for (ind = 0 ; ind < nbr_fac ; ind++)
    (*fac_val_reference)[ind] = renum[(*fac_val_reference)[ind]] ;

  BFT_FREE(renum) ;

  BFT_REALLOC(cpt_elt_coul_ent[ECS_ENTMAIL_FAC],
              cpt_coul_ent[ECS_ENTMAIL_FAC], ecs_size_t);

  elt_val_color_ent[ECS_ENTMAIL_FAC] = (*fac_val_reference) ;


  /* Transfert des valeurs lues dans les structures d'entite de maillage */
  /*=====================================================================*/


  vect_entmail = ecs_entmail_pre__cree_elt(nbr_elt_ent,
                                           elt_pos_som_ent,
                                           elt_val_som_ent,
                                           elt_val_typ_geo_ent,
                                           NULL,
                                           NULL,
                                           elt_val_color_ent,
                                           cpt_coul_ent,
                                           val_coul_ent,
                                           cpt_elt_coul_ent) ;

  vect_entmail[ECS_ENTMAIL_SOM] = vect_entmail_som ;

  /* Les tableaux suivants sont librs par ecs_entmail_pre__cree_elt() */

  *are_val_som = NULL ;
  *fac_val_som = NULL ;
  *cel_val_som = NULL ;
  *fac_val_reference = NULL ;


  /* Renvoi de la structure associe */

  return vect_entmail ;

}


/*----------------------------------------------------------------------------
 *   Transformation d'une hirarchie en filiation et affectation  une entit
 *----------------------------------------------------------------------------*/

static void ecs_loc_pre_hex__cree_filiation
(
 ecs_int_t               nbr_elt           , /* --> Nombre d'lments         */
 ecs_int_t      * *const ele_val_hierarch  , /* <-> Hirarchie des lments   */
 ecs_entmail_t    *const entmail             /* <-> Entit de maillage        */
)
{

  ecs_int_t     ind ;
  ecs_int_t     ind_fils ;
  ecs_int_t     ind_frere ;
  ecs_int_t     cpt_mod ;

  ecs_champ_t  *champ_filiation ;

  ecs_int_t    *filiation ;

  ecs_bool_t    propag_filiation ;

  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  /* Initialisation pre -> fils */

  propag_filiation = ECS_FALSE ;

  BFT_MALLOC(filiation, nbr_elt, ecs_int_t) ;

  for (ind = 0 ; ind < nbr_elt ; ind++)
    filiation[ind] = 0 ;

  for (ind = 0 ; ind < nbr_elt ; ind++) {

    ind_fils = (*ele_val_hierarch)[ind * 2] ;

    if (ind_fils != -1) {

      filiation[ind_fils] = ind + 1 ;

      if (propag_filiation == ECS_FALSE)
        propag_filiation = ECS_TRUE ;

    }

  }


  if (propag_filiation == ECS_FALSE) {

    BFT_FREE(*ele_val_hierarch) ;
    BFT_FREE(filiation) ;

    return ;

  }


  /* Propagation rcursive de frre en frre */

  do {

    cpt_mod = 0 ;

    for (ind = 0 ; ind < nbr_elt ; ind++) {

      if (filiation[ind] > 0) {

          ind_frere = (*ele_val_hierarch)[ind * 2 + 1] ;

          if (ind_frere > -1 && filiation[ind_frere] == 0) {
            filiation[ind_frere] = filiation[ind] ;
            cpt_mod++;
          }

      }

    }

  } while (cpt_mod > 0) ;


  BFT_FREE(*ele_val_hierarch) ;


  /* Branchement dans la structure entmail */
  /*---------------------------------------*/

  /* Transformation du tableau en une table */

  champ_filiation = ecs_champ__transforme_tableau(nbr_elt,
                                                  filiation,
                                                  ECS_CHAMP_NOM_FILIATION,
                                                  NULL,
                                                  ECS_CHAMP_STATUT_REF_ELT) ;

  BFT_FREE(filiation) ;


  /* Transfert dans l'entit de maillage */


  if (entmail != NULL && champ_filiation != NULL) {

    ecs_entmail__ajoute_champ(entmail,
                              champ_filiation,
                              ECS_CHAMP_ATT) ;

  }

}


/*----------------------------------------------------------------------------
 *   Lecture d'un entier au format " %d" sur fichier ASCII avec des lignes
 *  potentiellement trs longues. On lit alors une partie de la ligne
 *  dans un tampon, dont on fait glisser le contenu aprs extraction de
 *  l'entier.
 *----------------------------------------------------------------------------*/

static int ecs_loc_pre_hex__scan_d_ligne
(
 bft_file_t       *const fic               , /* --> Fichier                   */
 ecs_int_t        *const num_ligne         , /* <-> Numro de ligne           */
 char             *const chaine            , /* <-> Tampon associ            */
 ecs_int_t               len                 /* <-> Longueur de la chane     */
)
{

  ecs_int_t  p1, p2, res ;

  for (p1 = 0 ;
       chaine[p1] != '\0' && chaine[p1] != '0' && chaine[p1] != '1'
         && chaine[p1] != '2' && chaine[p1] != '3' && chaine[p1] != '4'
         && chaine[p1] != '5' && chaine[p1] != '6' && chaine[p1] != '7'
         && chaine[p1] != '8' && chaine[p1] != '9' ;
       p1++) {

    if (chaine[p1] == '\0')
      bft_file_gets(chaine, len, fic, num_ligne) ;

  }

  for (p2 = p1 + 1 ;
       chaine[p2] != '\0'  && chaine[p2] != '\n' && chaine[p2] != ' ' ;
       p2++) ;

  /* Si le tampon semble coup, on en relit une partie */
  if (chaine[p2] == '\0')
    bft_file_gets(chaine+p2, len-p2, fic, num_ligne) ;

  while (chaine[p2] != '\0' && chaine[p2] != '\n' && chaine[p2] != ' ')
    p2++ ;

  assert (chaine[p2] != '\0') ;

  if (sscanf(chaine+p1, "%d", &res) != 1)
    bft_error(__FILE__, __LINE__, errno,
              _("Error reading line %d of file \"%s\"."),
              *num_ligne, bft_file_get_name(fic));

  /* Dcalage du tampon */

  for (p1 = 0 ; chaine[p2] != '\0' ; p1++, p2++)
    chaine[p1] = chaine[p2] ;
  chaine[p1] = '\0';

  return res ;

}

