/*============================================================================
 *  Dfinitions des fonctions de base
 *   ralisant les sorties pour l'interfacage avec le Noyau
 *============================================================================*/

/*
                      Code_Saturne version 1.3
                      ------------------------

  This file is part of the Code_Saturne Preprocessor, element of the
  Code_Saturne CFD tool.

  Copyright (C) 1999-2007 EDF S.A., France

  contact: saturne-support@edf.fr

  The Code_Saturne Preprocessor is free software; you can redistribute it
  and/or modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2 of
  the License, or (at your option) any later version.

  The Code_Saturne Preprocessor is distributed in the hope that it will be
  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with the Code_Saturne Preprocessor; if not, write to the
  Free Software Foundation, Inc.,
  51 Franklin St, Fifth Floor,
  Boston, MA  02110-1301  USA
*/


/*============================================================================
 *                                 Visibilit
 *============================================================================*/

#include "ecs_config.h"


/*----------------------------------------------------------------------------
 *  Fichiers `include' librairie standard C ou BFT
 *----------------------------------------------------------------------------*/

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <bft_mem.h>
#include <bft_file.h>
#include <bft_error.h>
#include <bft_printf.h>


/*----------------------------------------------------------------------------
 *  Fichiers `include' visibles du  paquetage global "Utilitaire"
 *----------------------------------------------------------------------------*/

#include "ecs_chaine_glob.h"
#include "ecs_def.h"


/*----------------------------------------------------------------------------
 *  Fichier  `include' du  paquetage courant associ au fichier courant
 *----------------------------------------------------------------------------*/

#include "ecs_comm.h"


/*----------------------------------------------------------------------------
 *  Fichiers `include' prives du  paquetage courant
 *----------------------------------------------------------------------------*/

#include "ecs_comm_priv.h"


/*============================================================================
 *  Constantes et Macros
 *============================================================================*/


/*============================================================================
 *  Variables globales statiques
 *============================================================================*/


/*============================================================================
 *                       Prototypes de fonctions prives
 *============================================================================*/


/*----------------------------------------------------------------------------
 *  Fonction qui construit le descripteur du fichier d'interface et initialise
 *   ce fichier par l'envoi ou la lecture d'une ventuelle "chane magique"
 *   servant a vrifier le bon format des fichiers
 *----------------------------------------------------------------------------*/

static void ecs_loc_comm__fic_ouvre
(
       ecs_comm_t  *comm,
 const char        *nom_fic_comm
) ;


/*----------------------------------------------------------------------------
 *  Affichage de l'attente d'echange  d'un message
 *----------------------------------------------------------------------------*/

static void ecs_loc_comm__echo_pre
(
 const ecs_comm_t  *comm
) ;


/*----------------------------------------------------------------------------
 *  Affichage de l'entete d'une rubrique
 *
 *  1. nom     de la rubrique
 *  2. nombre          d'lments
 *  3. nom     du type d'lments
 *----------------------------------------------------------------------------*/

static void ecs_loc_comm__echo_entete
(
 const char    *nom_rub,
       size_t   nbr_elt_rub,
 const char    *nom_typ_e
) ;


/*----------------------------------------------------------------------------
 *  Affichage (partiel) du contenu d'une rubrique
 *----------------------------------------------------------------------------*/

static void ecs_loc_comm__echo_donnees
(
 const ecs_comm_t    *const comm,
       size_t               nbr_elt_rub,
       ecs_type_t     const typ_elt_e,
 const void          *const elt_rub
) ;


/*----------------------------------------------------------------------------
 *  Fonction qui crit un enregistrement dans le fichier d'interface
 *----------------------------------------------------------------------------*/

static void ecs_loc_comm__ecrit_rec
(
 const ecs_comm_t  *const comm,
 const void        *      rec,
       size_t             nbr,
       ecs_type_t         typ_e
) ;


/*============================================================================
 *                             Fonctions publiques
 *============================================================================*/

/*----------------------------------------------------------------------------
 *  Fonction qui initialise un communicateur
 *----------------------------------------------------------------------------*/

ecs_comm_t * ecs_comm__initialise
(
 ecs_int_t        num_noeud,                /* Numro du noeud associ        */
 ecs_comm_typ_t   type_comm,                /* Type de communicateur          */
 ecs_int_t        echo_comm                 /* Echo sur sortie principale     */
)
{

  ecs_comm_t  * comm ;

  const char  nom_base[] = "preprocessor_output" ;

  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  BFT_MALLOC(comm, 1, ecs_comm_t) ;

  /* Construction du nom du communicateur */
  /*--------------------------------------*/

  BFT_MALLOC(comm->nom, strlen(nom_base) + 1, char) ;

  sprintf(comm->nom, nom_base) ;

  /* Initialisation des autres champs */

  comm->type = type_comm ;

  comm->fic = NULL ;

  comm->echo = echo_comm ;

  comm->num_noeud = num_noeud ;


  /* Saut d'une ligne dans le listing */

  bft_printf("\n") ;


  /* Info sur la creation de l'interface */

  if (num_noeud == 0)
    bft_printf(_("  Opening communication: %s\n"),
               comm->nom) ;
  else
    bft_printf(_("  Opening communication: %s (node %4d)\n"),
               comm->nom, (int)num_noeud) ;

  bft_printf_flush();


  /* Creation du descripteur de fichier d'interface */
  /*------------------------------------------------*/

  if (comm->type == ECS_COMM_TYP_BINAIRE) {

    char * nom_fic_comm = NULL ;

    BFT_MALLOC(nom_fic_comm, strlen(nom_base) + 2 + 5 + 1, char) ;

    if (bft_file_mkdir_default(nom_base) == 0)
      sprintf(nom_fic_comm, "%s/n%05d", nom_base, num_noeud) ;

    else
      sprintf(nom_fic_comm, "%s_n%05d", nom_base, num_noeud) ;


    ecs_loc_comm__fic_ouvre(comm, nom_fic_comm) ;


    BFT_FREE(nom_fic_comm) ;

  }


  return comm ;

}


/*----------------------------------------------------------------------------
 *  Fonction qui termine la communication
 *----------------------------------------------------------------------------*/

ecs_comm_t * ecs_comm__termine
(
 ecs_comm_t  *comm
)
{

  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  if (comm->echo >= 0)
    bft_printf("\n") ;


  if (comm->num_noeud == 0)
    bft_printf(_("  Closing communication: %s\n"),
               comm->nom) ;

  else
    bft_printf(_("  Closing communication: %s (node %4d)\n"),
               comm->nom, (int)comm->num_noeud) ;


  if (comm->fic != NULL)
    comm->fic = bft_file_free(comm->fic) ;

  BFT_FREE(comm->nom) ;

  BFT_FREE(comm) ;


  return NULL ;

}


/*----------------------------------------------------------------------------
 *  Fonction d'interrogation du nom d'un communicateur
 *
 *  Cette fonction renvoie un pointeur sur un nom existant, et ne cre donc
 *   pas une nouvelle chane de caracteres. On ne doit donc pas dsallouer
 *   la chane pointe par cette fonction.
 *----------------------------------------------------------------------------*/

char * ecs_comm__ret_nom
(
 const ecs_comm_t  *comm
)
{
  return comm->nom ;
}


/*----------------------------------------------------------------------------
 * Fonction qui crit une rubrique dans le fichier d'interface
 *
 * Descriptif d'une rubrique du fichier d'interface avec le Noyau :
 *
 * 1. nom     de la rubrique
 * 2. nombre          d'lements
 * 3. nom     du type d'lements
 * 4. valeurs des       lements
 *----------------------------------------------------------------------------*/

void ecs_comm__ecrit_rubrique
(
 const char        *nom_rub,
       size_t       nbr_elt_rub,
 const void        *elt_rub,
       ecs_type_t   typ_elt_e,
 const ecs_comm_t  *comm
)
{

  char nom_rub_ecr[ECS_COMM_LNG_NOM_RUB + 1] ;
  char nom_typ_elt[2 + 1] ;


  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  assert(comm != NULL) ;


  /* nom de la rubrique */

  sprintf(nom_rub_ecr,
          "%-*.*s",
          ECS_COMM_LNG_NOM_RUB,
          ECS_COMM_LNG_NOM_RUB,
          nom_rub) ;


  /* nom du type d'elements */

  if (nbr_elt_rub != 0) {

    switch(typ_elt_e) {

    case ECS_TYPE_ecs_int_t:

      switch(sizeof(ecs_int_t)) {
      case 4:
        strcpy(nom_typ_elt, "i4") ;
        break ;
      case 8:
        strcpy(nom_typ_elt, "i8") ;
        break ;
      default:
        assert(sizeof(ecs_int_t) == 4 || sizeof(ecs_int_t) == 8) ;
      }
      break ;

    case ECS_TYPE_ecs_real_t:

      switch(sizeof(ecs_real_t)) {
      case 4:
        strcpy(nom_typ_elt, "r4") ;
        break ;
      case 8:
        strcpy(nom_typ_elt, "r8") ;
        break ;
      default:
        assert(sizeof(ecs_real_t) == 4 || sizeof(ecs_real_t) == 8) ;
      }
      break ;

    case ECS_TYPE_ecs_size_t:

      switch(sizeof(ecs_size_t)) {
      case 4:
        strcpy(nom_typ_elt, "u4") ;
        break ;
      case 8:
        strcpy(nom_typ_elt, "u8") ;
        break ;
      default:
        assert(sizeof(ecs_size_t) == 4 || sizeof(ecs_size_t) == 8) ;
      }
      break ;

    case ECS_TYPE_size_t:

      strcpy(nom_typ_elt, "u8") ;
      break ;

    case ECS_TYPE_char:

      strcpy(nom_typ_elt, "c ") ;
      break ;

    default:

      assert(typ_elt_e == ECS_TYPE_ecs_int_t   ||
             typ_elt_e == ECS_TYPE_ecs_real_t  ||
             typ_elt_e == ECS_TYPE_ecs_size_t  ||
             typ_elt_e == ECS_TYPE_size_t      ||
             typ_elt_e == ECS_TYPE_char        ) ;


    } /* Fin `switch (typ_elt_e)' */

  }


  if (comm->echo  >= 0)

    ecs_loc_comm__echo_pre(comm) ;


  /*------------------------------------------------------------------------*/
  /* Cas d'un fichier                                                       */
  /*------------------------------------------------------------------------*/

  if (comm->type == ECS_COMM_TYP_BINAIRE) {

    /* 1. nom     de type de la rubrique */
    /*-----------------------------------*/

    ecs_loc_comm__ecrit_rec(comm,
                            nom_rub_ecr,
                            ECS_COMM_LNG_NOM_RUB,
                            ECS_TYPE_char) ;

    /* 2. nombre          d'lments */
    /*-------------------------------*/

    ecs_loc_comm__ecrit_rec(comm,
                            &nbr_elt_rub,
                            1,
                            ECS_TYPE_size_t) ;


    if (nbr_elt_rub != 0) {

      /* 3. nom     du type d'lments */
      /*-------------------------------*/

      ecs_loc_comm__ecrit_rec(comm,
                              nom_typ_elt,
                              2,
                              ECS_TYPE_char) ;

      /* 4. valeurs des       lments */
      /*-------------------------------*/

      ecs_loc_comm__ecrit_rec(comm,
                              elt_rub,
                              nbr_elt_rub,
                              typ_elt_e) ;

    } /* Fin : s'il y a des lments  crire */

  }


  /*------------------------------------------------------------------------*/
  /* Affichage eventuel                                                     */
  /*------------------------------------------------------------------------*/

  if (comm->echo > 0) {

    ecs_loc_comm__echo_entete(nom_rub,
                              nbr_elt_rub,
                              nom_typ_elt) ;

    ecs_loc_comm__echo_donnees(comm,
                               nbr_elt_rub,
                               typ_elt_e,
                               elt_rub) ;

  }

}


/*============================================================================
 *                              Fonctions prives
 *============================================================================*/

/*----------------------------------------------------------------------------
 *  Fonction qui construit le descripteur du fichier d'interface et initialise
 *   ce fichier par l'envoi ou la lecture d'une eventuelle "chaine magique"
 *   servant a verifier le bon format des fichiers
 *----------------------------------------------------------------------------*/

static void ecs_loc_comm__fic_ouvre
(
       ecs_comm_t        *comm,
 const char       *const  nom_fic_comm
)
{

  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  /* Creation du descripteur du fichier d'interface */
  /*------------------------------------------------*/

  comm->fic = bft_file_open(nom_fic_comm,
                            BFT_FILE_MODE_WRITE,
                            BFT_FILE_TYPE_BINARY) ;

  bft_file_set_big_endian(comm->fic) ;


  /* criture d'une "chane magique" */
  /*---------------------------------*/

  ecs_loc_comm__ecrit_rec(comm,
                          "ECS_1.3",
                          7, /* longueur de la chane */
                          ECS_TYPE_char) ;

}


/*----------------------------------------------------------------------------
 *  Affichage de l'attente d'echange  d'un message
 *----------------------------------------------------------------------------*/

static void ecs_loc_comm__echo_pre
(
 const ecs_comm_t      *const comm
)
{

  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  assert(comm != NULL) ;


  if (comm->num_noeud == 0)
    bft_printf(_("\nSection written on \"%s\" :\n"), comm->nom) ;
  else
    bft_printf(_("\nSection written on \"%s\" (node %d) :\n"), comm->nom,
               comm->num_noeud) ;

  bft_printf_flush() ;

}


/*----------------------------------------------------------------------------
 *  Affichage de l'entte d'une rubrique
 *
 *  1. nom     de la rubrique
 *  2. nombre          d'lments
 *  3. nom     du type d'lments
 *----------------------------------------------------------------------------*/

static void ecs_loc_comm__echo_entete
(
 const char    *nom_rub,
       size_t   nbr_elt_rub,
 const char    *nom_typ_e
)
{

  char nom_rub_ecr[ECS_COMM_LNG_NOM_RUB + 1] ;


  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  strncpy(nom_rub_ecr, nom_rub,   ECS_COMM_LNG_NOM_RUB) ;
  nom_rub_ecr[ECS_COMM_LNG_NOM_RUB] = '\0' ;

  bft_printf(_("    section name          : \"%s\"\n"
               "    number of elements    : %lu \n"),
             nom_rub_ecr, (unsigned long)nbr_elt_rub) ;


  if (nbr_elt_rub > 0) {

    char nom_typ_ecr[ECS_COMM_LNG_NOM_TYP_ELT + 1] ;

    strncpy(nom_typ_ecr, nom_typ_e, ECS_COMM_LNG_NOM_TYP_ELT) ;
    nom_typ_ecr[ECS_COMM_LNG_NOM_TYP_ELT] = '\0' ;

    bft_printf(_("    element type name     : \"%s\"\n"), nom_typ_ecr) ;

  }


  bft_printf_flush() ;

}


/*----------------------------------------------------------------------------
 *  Affichage (partiel) du contenu d'une rubrique
 *----------------------------------------------------------------------------*/

static void ecs_loc_comm__echo_donnees
(
 const ecs_comm_t    *comm,
       size_t         nbr_elt_rub,
       ecs_type_t     typ_elt_e,
 const void          *elt_rub
)
{

  size_t  echo_deb ;
  size_t  echo_fin ;
  size_t  ind ;

  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/


  if (nbr_elt_rub == 0) return ;

  assert(comm != NULL) ;

  echo_deb = 0 ;

  if (comm->echo * 2 < (ecs_int_t)nbr_elt_rub) {

    echo_fin = comm->echo ;

    bft_printf(_("    %d first and last elements:\n"), comm->echo) ;

  }
  else {

    echo_fin = nbr_elt_rub ;

    bft_printf(_("    elements:\n")) ;

  }


  do {

    switch(typ_elt_e) {

    case ECS_TYPE_ecs_int_t: {

      const ecs_int_t *elt_rub_int = (const ecs_int_t *) elt_rub ;

      for (ind = echo_deb ; ind < echo_fin ; ind++)
        bft_printf("    %10lu : %12d\n",
                   (unsigned long)(ind + 1), *(elt_rub_int + ind)) ;

    }
    break ;

    case ECS_TYPE_ecs_real_t: {

      const ecs_real_t *elt_rub_real = (const ecs_real_t *) elt_rub ;

      for (ind = echo_deb ; ind < echo_fin ; ind++)
        bft_printf("    %10lu : %12.5e\n",
                   (unsigned long)(ind + 1), *(elt_rub_real + ind)) ;

    }
    break ;

    case ECS_TYPE_ecs_size_t: {

      const ecs_size_t *elt_rub_int = (const ecs_size_t *) elt_rub ;

      for (ind = echo_deb ; ind < echo_fin ; ind++)
        bft_printf("    %10lu : %12d\n",
                   (unsigned long)(ind + 1), *(elt_rub_int + ind)) ;

    }
    break ;

    case ECS_TYPE_size_t: {

      const size_t *elt_rub_int = (const size_t *) elt_rub ;

      for (ind = echo_deb ; ind < echo_fin ; ind++)
        bft_printf("    %10lu : %12d\n",
                   (unsigned long)(ind + 1), *(elt_rub_int + ind)) ;

    }
    break ;

    case ECS_TYPE_char: {

      const char *elt_rub_char = (const char *) elt_rub ;

      for (ind = echo_deb ; ind < echo_fin ; ind++) {
        if (*(elt_rub_char + ind) != '\0')
          bft_printf("    %10lu : '%c'\n",
                     (unsigned long)(ind + 1), *(elt_rub_char + ind)) ;
        else
          bft_printf("    %10lu : '\\0'\n", (unsigned long)(ind + 1)) ;
      }

    }
    break ;

    default:

      assert(typ_elt_e == ECS_TYPE_ecs_int_t  ||
             typ_elt_e == ECS_TYPE_ecs_real_t ||
             typ_elt_e == ECS_TYPE_ecs_size_t ||
             typ_elt_e == ECS_TYPE_size_t     ||
             typ_elt_e == ECS_TYPE_char         ) ;

    }


    if (echo_fin < nbr_elt_rub) {

      bft_printf(_("    ..........   ............\n")) ;

      echo_deb = nbr_elt_rub - comm->echo ;
      echo_fin = nbr_elt_rub ;

    }
    else {

      assert(echo_fin == nbr_elt_rub) ;

      echo_fin = nbr_elt_rub + 1 ;

    }


  } while (echo_fin <= nbr_elt_rub) ;


  bft_printf_flush() ;

}


/*----------------------------------------------------------------------------
 *  Fonction qui crit un enregistrement dans le fichier d'interface
 *
 *  Cette fonction slectionne le mode d'criture de l'enregistrement
 *  suivant le type du fichier
 *----------------------------------------------------------------------------*/

static void ecs_loc_comm__ecrit_rec
(
 const ecs_comm_t  *comm,
 const void        *rec,
       size_t       nbr,
       ecs_type_t   typ_e
)
{

  size_t taille ;

  /*xxxxxxxxxxxxxxxxxxxxxxxxxxx Instructions xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx*/

  assert(comm != NULL) ;
  assert(rec  != NULL) ;


  /* Dtermination de la taille associe au type  envoyer */

  switch (typ_e) {
  case ECS_TYPE_ecs_int_t:
    taille = sizeof(ecs_int_t) ;
    break ;
  case ECS_TYPE_ecs_real_t:
    taille = sizeof(ecs_real_t) ;
    break ;
  case ECS_TYPE_ecs_size_t:
    taille = sizeof(ecs_size_t) ;
    break ;
  case ECS_TYPE_char:
    taille = sizeof(char) ;
    break ;
  case ECS_TYPE_size_t:
    taille = 8 ;
    break ;
  default:
    assert(typ_e == ECS_TYPE_ecs_int_t  ||
           typ_e == ECS_TYPE_ecs_real_t ||
           typ_e == ECS_TYPE_ecs_size_t ||
           typ_e == ECS_TYPE_size_t     ||
           typ_e == ECS_TYPE_char ) ;
  }

  /* criture de l'enregistrement dans le fichier */
  /*----------------------------------------------*/

  if (typ_e != ECS_TYPE_size_t || sizeof(size_t) == 8)
    bft_file_write(rec,
                   taille,
                   nbr,
                   comm->fic) ;

  else { /* if (typ_e == ECS_TYPE_size_t && sizeof(size_t) != 8) */

    unsigned long long _nbr = nbr ;

    assert(sizeof(unsigned long long) == 8) ;

    bft_file_write(rec,
                   taille,
                   _nbr,
                   comm->fic) ;

  }

}

